(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Bool List Arith Omega Relations.
Require Import vgtac myclassical.
Set Implicit Arguments.



(** * Random stuff. *)

Arguments Basics.compose [A B C] g f x /.

Coercion sumbool_to_bool A B (x: sumbool A B) : bool :=
  if x then true else false.

Lemma eta: forall {A B} (f: A -> B),
  f = fun x => f x.
Proof.
  auto.
Qed.

Lemma prod_eta: forall X Y (p: X * Y),
  p = (fst p, snd p).
Proof.
  i; destruct p; auto.
Qed.

Lemma eq_fun: forall A B (f g: A -> B)
    (E: f = g),
  forall x, f x = g x.
Proof.
  i; rewrite E; auto.
Qed.

Lemma change_arg: forall X (f: X -> Prop) x x',
  f x' -> x' = x -> f x.
Proof.
  i; rewrite H0 in *; auto.
Qed.

Definition option_map2 {A B C} (f: A -> B -> C) (x: option A) (y: option B) : option C :=
  match x,y with
  | Some a, Some b => Some (f a b)
  | _,_ => None
  end.

(** Turn [option A] into [A], using a default value. *)
Definition withd A (a: A) (o: option A) : A :=
  match o with
    | None => a
    | Some a' => a'
  end.


Definition case X Y Z (f: X -> Z) (g: Y -> Z) : X + Y -> Z :=
  sum_rect (fun _ => _) f g.

Definition Inl X Y Z (f: X -> Z) : X + Y -> Z + Y :=
  case (inl <*> f) inr.

Definition Inr X Y Z (f: Y -> Z) : X + Y -> X + Z :=
  case inl (inr <*> f).



(** * Type class for types with decidable equality. *)

Class EqDec (A: Type) : Type := mkEqDec {
  eqdec: forall a b : A, {a = b} + {a <> b}
}.

Lemma eqdec_true X {M:EqDec X}: forall x,
  eqdec x x = left _ refl_equal.
Proof.
  by i; destruct eqdec; [f_equal; apply UIP|].
Qed.

Lemma eqdec_false X {M:EqDec X}: forall x y
    (NEQ: x <> y),
  eqdec x y = right _ NEQ.
Proof.
  by i; destruct eqdec; [exfalso; eauto|f_equal; apply proof_irrelevance].
Qed.


(** ** Instance for [nat]. *)
Instance NatEqDec: EqDec nat :=
  mkEqDec eq_nat_dec.


(** ** Instance lifting for products. *)
Section SumEqDec.
  Variable A : Type.
  Variable EA : EqDec A.
  Variable B: Type.
  Variable EB : EqDec B.

  Definition eq_sum_dec (x y: A + B) : {x = y} + {x <> y}.
    destruct x, y.
    - destruct (eqdec a a0); subst; auto; right; ii; vauto.
    - right; vauto.
    - right; vauto.
    - destruct (eqdec b b0); subst; auto; right; ii; vauto.
  Defined.
End SumEqDec.

Instance SumEqDec A (EA: EqDec A) B (EB: EqDec B) : EqDec (sum A B) :=
  mkEqDec (eq_sum_dec EA EB).



(** * Additions to the Bool library. *)


Lemma Bool_leb_refl: forall x,
  Bool.leb x x.
Proof.
  i; destruct x; s; auto.
Qed.
Hint Resolve Bool_leb_refl.

Lemma Bool_leb_trans: forall
    x y (LE: Bool.leb x y)
    z (LE': Bool.leb y z),
  Bool.leb x z.
Proof.
  i; destruct x, y, z; s; auto.
Qed.



(** * Additions to the List library. *)

Lemma lt_length_well_founded: forall {A},
  well_founded (fun xs ys : list A => length xs < length ys).
Proof.
  ii; remember (length a) as n; depgen a.
  induction n using (well_founded_ind lt_wf).
  econs; i; subst; eauto.
Qed.

Section ListDefs.

Variables A B C D : Type.
Variable myeq : forall a b:A,{a=b}+{a<>b}. 
Instance myEqDec: EqDec A :=
  mkEqDec myeq.

Implicit Types (a : A) (b : B).

(** ** New functions and relations. *)

Definition last_error (l: list A) :=
  last (map (@Some _) l) None.

Fixpoint repeat (x: A) n :=
  match n with
    | O => nil
    | S n' => x :: repeat x n'
  end.

Fixpoint list_index X {M: EqDec X} (x: X) l :=
  match l with
    | nil => None
    | y :: l' =>
      if eqdec x y then Some 0 else option_map S (list_index x l')
  end.

Definition map_pair (f : A -> B) (g : C -> D) l :=
  map (fun x => (f (fst x), g (snd x))) l.

Fixpoint list_mem (x:A) (l:list A) : bool :=
  match l with
    | nil => false
    | cons h t => if myeq h x then true else list_mem x t
  end.

Fixpoint list_assoc (x:A) (l:list (A*B)) : option B :=
  match l with
    | nil => None
    | cons (a,b) t => if myeq a x then Some b else list_assoc x t
  end.

Fixpoint list_minus (l1:list A) (l2:list A) : list A :=
  match l1 with
    | nil => nil
    | cons h t => if list_mem h l2 then list_minus t l2 else cons h (list_minus t l2)
  end.

Fixpoint list_minus2 (l1:list (A*B)) (l2:list A) : list (A*B) :=
  match l1 with
    | nil => nil
    | cons h t => if list_mem (fst h) l2 then list_minus2 t l2 else cons h (list_minus2 t l2)
  end.

Fixpoint list_disj (l l': list A) : Prop :=
  match l' with
    | nil => True
    | cons h t => if list_mem h l then False else list_disj l t
  end.

Fixpoint list_remdup (l : list A) : list A :=
  match l with
    | nil => nil
    | h :: t => h :: list_minus (list_remdup t) (cons h nil)
  end.

Fixpoint list_remdup2 (l : list (A * B)) : list (A * B) :=
  match l with
    | nil => nil
    | (x,e) :: l => (x, e) :: list_minus2 (list_remdup2 l) (cons x nil)
  end.

Definition list_max : list nat -> nat :=
  fold_right max 0.

Inductive sublist X {M: EqDec X}: list X -> list X -> Prop :=
| sublist_nil:
    sublist nil nil
| sublist_cons_drop: forall x l l'
      (SB: sublist l l'),
    sublist l (x::l')
| sublist_cons_keep: forall x l l'
      (SB: sublist l l'),
    sublist (x::l) (x::l')
.
Hint Constructors sublist.

Fixpoint combine_full X Y (xs: list X) (ys: list Y) : list (X * Y) :=
  match xs with
    | nil => nil
    | x :: xs' => map (fun y => (x, y)) ys ++ combine_full xs' ys
  end.

Fixpoint flatten X (l: list (list X)) :=
  match l with
   | nil => nil
   | x :: l' => x ++ flatten l'
  end.

(** ** Properties of these new definitions as well as of some old ones. *)


(** *** Properties of [cons]. *)

Lemma cons_app: forall X x1 (xs: list X),
  x1::xs = [x1] ++ xs.
Proof. trivial. Qed.
Lemma cons_app2: forall X x1 x2 (xs: list X),
  x1::x2::xs = [x1; x2] ++ xs.
Proof. trivial. Qed.
Lemma cons_app3: forall X x1 x2 x3 (xs: list X),
  x1::x2::x3::xs = [x1; x2; x3] ++ xs.
Proof. trivial. Qed.
Lemma cons_app4: forall X x1 x2 x3 x4 (xs: list X),
  x1::x2::x3::x4::xs = [x1; x2; x3; x4] ++ xs.
Proof. trivial. Qed.
Lemma cons_app5: forall X x1 x2 x3 x4 x5 (xs: list X),
  x1::x2::x3::x4::x5::xs = [x1; x2; x3; x4; x5] ++ xs.
Proof. trivial. Qed.


(** *** Properties of [combine]. *)

Lemma combine_fst_snd: forall X Y (ps: list (X * Y)),
  combine (map fst ps) (map snd ps) = ps.
Proof.
  induction ps; auto.
  destruct a; s; f_equal; eauto.
Qed.


(** *** Properties of [flatten]. *)

Lemma flatten_app: forall X (l1: list (list X)) l2,
  flatten (l1 ++ l2) = flatten l1 ++ flatten l2.
Proof.
  induction l1; auto.
  s; i; rewrite <-app_assoc, IHl1; eauto.
Qed.

Lemma flatten_inD: forall X (x: X) l
    (IN: In x (flatten l)),
  exists y, <<INa: In x y>> /\ <<INb: In y l>>.
Proof.
  i; induction l; vauto.
  ss; apply in_app_or in IN; des; eauto.
  apply IHl in IN; des; eauto.
Qed.


(** *** Properties of [nth_error]. *)

Lemma nth_error_length: forall X n xs (x: X)
    (IN: nth_error xs n = Some x),
  n < length xs.
Proof.
  induction n, xs; try (intros; inversion IN).
  - unfold nth_error; simpl in *; auto with arith.
  - intros; simpl in IN; apply lt_n_S; eauto.
Qed.

Lemma nth_error_app: forall X (la lb: list X) n x
    (IN: nth_error (la ++ lb) n = Some x),
  nth_error la n = Some x \/
  length la <= n /\ nth_error lb (n - length la) = Some x.
Proof.
  induction la; i; ss; [by rewrite <-minus_n_O; eauto using le_0_n|].
  destruct n; ss; eauto; exploit IHla; eauto; i; des; eauto using le_n_S.
Qed.

Lemma nth_error_some: forall X (x: X) l
    (IN: In x l),
  exists n, nth_error l n = Some x.
Proof.
  induction l; i; des; ss; des; vauto.
  - exists 0; eauto.
  - eapply IHl in IN; des; exists (S n); eauto.
Qed.

Lemma nth_error_some': forall n X (xs: list X)
    (LEN: n < length xs),
  exists x, nth_error xs n = Some x.
Proof.
  clear.
  i; ginduction xs; i; [by exfalso; ss; omega|].
  ss; destruct n; [by vauto|].
  apply lt_S_n in LEN; hexploit IHxs; eauto.
Qed.

Lemma nth_error_someD: forall X (x: X) l n
    (NTH: nth_error l n = Some x),
  In x l.
Proof.
  induction l; i; des; ss; [destruct n|]; des; vauto.
  destruct n; ss; clarify; inversion NTH; eauto.
Qed.


(** *** Properties of [map]. *)

Lemma map_cons: forall X Y (f: X -> Y) x xs,
  map f (x::xs) = f x :: map f xs.
Proof.
  auto.
Qed.

Lemma map_comp: forall X Y Z (f: Y -> Z) (g: X -> Y) xs,
  map (f <*> g) xs = map f (map g xs).
Proof.
  i; rewrite map_map; auto.
Qed.









(** *** Properties of [map_pair]. *)

Lemma map_fst_map_pair: forall f g l,
  map (@fst _ _) (map_pair f g l) = map f (map (@fst _ _) l).
Proof.
  unfold map_pair; intros; rewrite !map_map; eauto.
Qed.

Lemma map_snd_map_pair: forall f g l,
  map (@snd _ _) (map_pair f g l) = map g (map (@snd _ _) l).
Proof.
  unfold map_pair; intros; rewrite !map_map; eauto.
Qed.

Lemma map_pair_app: forall f g xs ys,
  map_pair f g (xs ++ ys) = map_pair f g xs ++ map_pair f g ys.
Proof.
  i; ginduction xs; auto.
  i; s; f_equal; eauto.
Qed.


(** *** Properties of [In]. *)

Lemma in_consD: forall X x y (l: list X)
    (IN: In x (y::l)),
  x = y \/ (In x l /\ x <> y).
Proof.
  i; destruct (LEM (x = y)); ss; des; vauto.
Qed.

Lemma in_split_first: forall x l,
  In x l -> exists l1 l2 : list A, l = l1 ++ x :: l2 /\ ~ In x l1.
Proof.
  i; induction l; vauto.
  apply in_consD in H; des; subst.
  - exists (@nil A), l; eauto.
  - apply IHl in H; des; subst.
    exists (a::l1), l2; split; s; auto.
    i; des; auto.
Qed.

Lemma nin_consD: forall X x y (l: list X)
    (IN: ~ In x (y::l)),
  y <> x /\ ~ In x l.
Proof.
  i; destruct (LEM (x = y)); splits; ss; des; auto; vauto.
Qed.

Lemma in_app_xor: forall (A : Type) (l m : list A) (a : A),
       In a (l ++ m) -> In a l \/ (~ In a l /\ In a m).
Proof.
  i; induction l; auto.
  rewrite <-app_comm_cons in H.
  apply in_consD in H.
  ss; des; subst; eauto.
  specialize (IHl H); des; [left|right]; try split; eauto.
  i; des; eauto.
Qed.

Lemma in_map_fst: forall x (l: list (A * B)), 
  In x (map fst l) <-> exists b, In (x, b) l.
Proof.
  induction l as [|[a b]]; simpl; i.
  - split; i; des; vauto.
  - split; i; des; subst; clarify; eauto.
    hexploit IHl; eauto; []; i.
    hexploit IHl0; eauto; []; i; des; eauto.
Qed.

Lemma in_map_snd: forall x (l: list (A * B)), 
  In x (map (@snd _ _) l) <-> exists a, In (a, x) l.
Proof.
  induction l as [|[a b]]; simpl; i.
  - split; i; des; vauto.
  - split; i; des; subst; clarify; eauto.
    hexploit IHl; eauto; []; i.
    hexploit IHl0; eauto; []; i; des; eauto.
Qed.


(** *** Properties of [list_mem]. *)

Lemma list_mem_app: forall a l1 l2,
  list_mem a (app l1 l2) = (list_mem a l1 || list_mem a l2).
Proof.
  induction l1 as [|a1 l1 IH]; intros; simpl; trivial; rewrite IH. 
  destruct myeq; trivial.
Qed.

Lemma list_mem_minus: forall a l1 l2,
  list_mem a (list_minus l1 l2) = (list_mem a l1 && negb (list_mem a l2)).
Proof.
  induction l1 as [|a1 l1 IH]; intros; simpl; trivial.
  destruct (list_mem a1 l2) as [] eqn: X; simpl; rewrite IH;
  destruct (myeq a1 a); subst; rewrite ?X; simpl; trivial using andb_false_r.
Qed.

Lemma list_mem_remdup: forall x l,
  list_mem x (list_remdup l) = list_mem x l.
Proof. 
  induction l as [|y l IH]; simpl in *; trivial.
  destruct (myeq y x); simpl; trivial; rewrite <- IH.
  generalize (list_remdup l); clear l IH.
  induction l as [|a l IH]; simpl in *; trivial. 
  destruct (myeq y); simpl; subst; destruct (myeq a); simpl; subst; tauto.
Qed.

Lemma list_mem_spec: forall x l,
  if list_mem x l then In x l else ~ In x l.
Proof.
  induction l as [|a l IH]; simpl; [|destruct (myeq a x)]; auto.
  destruct (list_mem); tauto. 
Qed.

Lemma list_mem_eq: forall x l1 l2,
  (In x l1 <-> In x l2) -> list_mem x l1 = list_mem x l2.
Proof.
  intros; generalize (list_mem_spec x l1), (list_mem_spec x l2).
  do 2 destruct (list_mem); tauto.
Qed.

Lemma list_mem_map_fst: forall x l, 
  list_mem x (map (@fst _ _) l) = 
  (match list_assoc x l with Some _ => true | None => false end).
Proof.
  induction l as [|[a b] l IH]; simpl; [|destruct (myeq a x)]; trivial.
Qed.

Lemma list_mem_false: forall x l,
  list_mem x l = false <-> ~ In x l.
Proof.
  i; assert (H:=list_mem_spec x l); des_ifH H; intuition.
Qed.

Lemma list_mem_true: forall x l,
  list_mem x l = true <-> In x l.
Proof.
  i; assert (H:=list_mem_spec x l); des_ifH H; intuition.
Qed.


(** *** Properties of [list_assoc]. *)

Lemma list_assoc_app: forall x l1 l2,
  list_assoc x (l1 ++ l2) =
  (match list_assoc x l1 with Some y => Some y | None => list_assoc x l2 end).
Proof.
  induction l1 as [|[a b] l1 IH]; intros; simpl; trivial; rewrite IH.
  destruct myeq; trivial. 
Qed.

Lemma list_assoc_app_or: forall x y l1 l2
    (IN: list_assoc x (l1 ++ l2) = Some y),
  list_assoc x l1 = Some y \/
  list_assoc x l1 = None /\ list_assoc x l2 = Some y.
Proof.
  i; rewrite list_assoc_app in IN.
  destruct (list_assoc x l1); auto.
Qed.

Lemma list_assoc_map: forall x l
    (IN: In x (map fst l)),
  exists y, list_assoc x l = Some y.
Proof.
  induction l; vauto; i; destruct a; s; des_if; eauto.
  ss; des; vauto; apply IHl; eauto.
Qed.

Lemma list_assoc_minus2: forall a l1 l2,
  list_assoc a (list_minus2 l1 l2) =
  if list_mem a l2 then None else list_assoc a l1.
Proof.
  induction l1 as [|[a1 b1] l1 IH]; i; s; [destruct list_mem; reflexivity|].
  destruct (list_mem a1 l2) as [] eqn: X; simpl; rewrite IH;
  destruct (myeq a1 a); subst; rewrite ?X; s; trivial using andb_false_r.
Qed.

Lemma list_assoc_remdup2: forall x l,
  list_assoc x (list_remdup2 l) = list_assoc x l.
Proof. 
  induction l as [|[y e] l IH]; simpl in *; trivial.
  destruct (myeq y x); simpl; trivial; rewrite <- IH.
  generalize (list_remdup2 l); clear l IH.
  induction l as [|[a b] l IH]; simpl in *; trivial. 
  destruct (myeq y); simpl; subst; destruct (myeq a); simpl; subst; tauto.
Qed.

Lemma list_assoc_spec: forall x l,
  match list_assoc x l with
    | None => ~ In x (map (@fst _ _) l)
    | Some y => In (x, y) l
  end.
Proof.
  induction l as [|[a b] l IH]; simpl; try tauto.
  destruct (myeq a x); subst; try tauto. 
  destruct (list_assoc x l); tauto. 
Qed.

Lemma list_assoc_noneD: forall x l
    (E: list_assoc x l = None),
  ~ In x (map (@fst A B) l).
Proof.
  intros; generalize (list_assoc_spec x l); rewrite E; auto.
Qed.

Lemma list_assoc_someD: forall x l y
    (E: list_assoc x l = Some y),
  In (x, y) l.
Proof.
  intros; generalize (list_assoc_spec x l); rewrite E; trivial.
Qed.

Lemma list_assoc_cons_eq: forall x y l,
  list_assoc x ((x, y) :: l) = Some y.
Proof.
  i; unfold list_assoc; des_if; vauto.
Qed.

Lemma list_assoc_cons_neq: forall x x' y l
    (NEQ: x <> x'),
  list_assoc x' ((x, y) :: l) = list_assoc x' l.
Proof.
  i; unfold list_assoc; des_if; vauto.
Qed.

Lemma list_assoc_cons_or: forall x x' y y' l
    (IN: list_assoc x ((x', y') :: l) = Some y),
  x = x' /\ y = y' \/
  x <> x' /\ list_assoc x l = Some y.
Proof.
  i; ss; des_ifH IN; vauto; auto.
Qed.

Lemma list_assoc_sng: forall x x' y y'
    (IN: list_assoc x [(x', y')] = Some y),
  x = x' /\ y = y'.
Proof.
  i; apply list_assoc_someD in IN; ss; des; vauto.
Qed.

Lemma list_assoc_cons_fresh: forall x y l x' y'
    (LU: list_assoc x l = Some y)
    (DJ: forall y' (LU: list_assoc x' l = Some y'), False),
  list_assoc x ((x', y') :: l) = Some y.
Proof.
  i.
  destruct (LEM (x = x')); subst.
  - exfalso; eapply DJ; eauto.
  - rewrite list_assoc_cons_neq; eauto.
Qed.


(** *** Properties of [list_minus]. *)

Lemma list_minus_nil: forall (xs: list A),
  list_minus xs xs = nil.
Proof.
  clear B C D.
  cut (forall (xs ys: list A), list_minus xs (ys++xs) = nil)
  ; i; [by specialize (H xs nil); eauto|]; ginduction xs; i; [by auto|].
  s; des_ifs.
  - rewrite cons_app, app_assoc; auto.
  - exfalso; rewrite list_mem_app in Heq; ss; des_ifs.
    destruct (list_mem _ _) in Heq; vauto.
Qed.

Lemma list_minus_notin: forall x xs
    (NIN: ~ In x xs),
  list_minus xs [x] = xs.
Proof.
  induction xs; [by auto|]; i; s; des_ifs.
  - exfalso; apply NIN; vauto.
  - rewrite IHxs; eauto; i; ss; eauto.
Qed.

Lemma in_list_minusI: forall x l l',
  In x l -> list_mem x l' = false -> In x (list_minus l l').
Proof.
  induction l as [|a l IH]; simpl; try tauto.
  intros l' [-> | IN] H; [rewrite H|destruct (list_mem a l')]; simpl; auto.
Qed. 

Lemma in_list_minus: forall x l l',
  In x (list_minus l l') <-> (In x l /\ list_mem x l' = false).
Proof.
  intros; induction l as [|a l]; simpl; [|destruct (list_mem a l') as [] eqn:?]; 
  simpl in *; intuition (subst; simpl in *; try congruence).
Qed.

Lemma in_list_minus2I: forall x l l',
  In x l -> list_mem (fst x) l' = false -> In x (list_minus2 l l').
Proof.
  induction l as [|a l IH]; simpl; try tauto.
  intros l' [-> | IN] H; [rewrite H|destruct (list_mem (fst a) l')]; simpl; auto.
Qed. 

Lemma in_list_minus2: forall x l l',
  In x (list_minus2 l l') <-> (In x l /\ list_mem (fst x) l' = false).
Proof.
  i; induction l as [|[a b] l]; s; [|destruct (list_mem a l') as [] eqn:?]
  ; ss; intuition (subst; ss; try congruence).
Qed.

Lemma list_minus_app: forall l1 l2 l,
  list_minus (l1 ++ l2) l = list_minus l1 l ++ list_minus l2 l.
Proof.
  induction l1; i; s; [|destruct (list_mem a l)]; s; f_equal; trivial.
Qed.

Lemma list_minus_app_r: forall l l1 l2,
  list_minus l (l1 ++ l2) = list_minus (list_minus l l1) l2.
Proof.
  induction l; i; s; try rewrite list_mem_app, IHl; trivial.
  do 2 (destruct (list_mem); simpl; trivial).
Qed.

Lemma list_minus_comm: forall l l1 l2,
  list_minus (list_minus l l1) l2 = list_minus (list_minus l l2) l1.
Proof.
  induction l as [|a l IH]; intros; simpl; trivial.
  destruct (list_mem a l1) as [] eqn: X; simpl;
  destruct (list_mem a l2) as [] eqn: Y; simpl; rewrite ?X, ?Y, IH; trivial.
Qed.

Lemma list_minus_map_fst: forall l l',
  list_minus (map (@fst _ _) l) l' = map (@fst A B) (list_minus2 l l').
Proof.
  induction l; simpl; intros; auto.
  destruct (list_mem); simpl; rewrite IHl; auto.
Qed.

Lemma list_minus_red: forall l1 l2 l,
  list_minus (list_minus l1 l) (list_minus l2 l) =
  list_minus (list_minus l1 l) l2.
Proof.
  induction l1 as [|a l1 IH]; simpl; intros; auto.
  destruct (list_mem a l) as [] eqn: X; simpl; auto.
  rewrite list_mem_minus, X, IH; destruct (list_mem a l2); trivial.
Qed.

Lemma list_minus_swap: forall l x y l',
  list_minus l (x :: y :: l') = list_minus l (y :: x :: l').
Proof.
  intros; induction l as [|a l IH]; simpl in *; trivial.
  rewrite IH; destruct (myeq x); destruct (myeq y); trivial.
Qed.


(** *** Properties of [list_minus2]. *)

Lemma list_minus2_app: forall l1 l2 l,
  list_minus2 (l1 ++ l2) l = list_minus2 l1 l ++ list_minus2 l2 l.
Proof.
  induction l1; i; s; [|destruct (list_mem _ _)]; s; f_equal; trivial.
Qed.

Lemma list_minus2_app_r: forall l l1 l2,
  list_minus2 l (l1 ++ l2) = list_minus2 (list_minus2 l l1) l2.
Proof.
  induction l; intros; simpl; try rewrite list_mem_app, IHl; trivial.
  do 2 (destruct (list_mem); simpl; trivial).
Qed.

Lemma list_minus2_comm: forall l l1 l2,
  list_minus2 (list_minus2 l l1) l2 = list_minus2 (list_minus2 l l2) l1.
Proof.
  induction l as [|[a b] l IH]; intros; simpl; trivial.
  destruct (list_mem a l1) as [] eqn: X; simpl;
  destruct (list_mem a l2) as [] eqn: Y; simpl; rewrite ?X, ?Y, IH; trivial.
Qed.

Lemma list_minus2_remdup2: forall l l',
  list_minus2 (list_remdup2 l) l' = list_remdup2 (list_minus2 l l').
Proof.
  induction l as [|[y e] l IH]; intros; simpl in *; trivial. 
  rewrite list_minus2_comm, IH. 
  case_eq (list_mem y l'); simpl; [f_equal|]; intros; simpl in *; trivial.
  rewrite <- IH; generalize (list_remdup2 l); clear l IH.
  induction l as [|[a b] l IH]; intros; simpl in *; trivial.
  case_eq (list_mem a l'); intros; simpl in *; trivial.
  rewrite IH; destruct (myeq y); subst; trivial; congruence.
Qed.

Lemma list_map_fst_minus2: forall l l',
  map (@fst A B) (list_minus2 l l') = list_minus (map (@fst _ _) l) l'.
Proof.
  symmetry; apply list_minus_map_fst.
Qed.

Lemma list_minus2_eq: forall l l1 l2
    (IN: forall x , In x l1 <-> In x l2),
  list_minus2 l l1 = list_minus2 l l2.
Proof.
  i; induction l; s; try rewrite IHl, (list_mem_eq _ _ _ (IN _)); auto.
Qed.


(** *** Properties of [list_disj]. *)

Lemma list_disj_cons_l: forall a l l',
  list_disj (a :: l) l' <-> (list_disj l l' /\ list_mem a l' = false).
Proof.
  induction l' as [|a' l' IH]; simpl; try tauto.
  destruct (myeq a a'); subst; try tauto; destruct (myeq a'); subst;
  do 2 destruct (list_mem); intuition. 
Qed.

Lemma list_disj_comm: forall l' l,
  list_disj l l' -> list_disj l' l.
Proof.
  induction l' as [|a l' IH]; simpl; auto.
  - solve [clear; induction l; simpl; auto].
  - intros; apply list_disj_cons_l; destruct (list_mem a l); intuition auto. 
Qed.

Lemma list_disj_app_rI: forall l l1 l2,
  list_disj l l1 -> list_disj l l2 -> list_disj l (l1 ++ l2).
Proof.
  induction l1; simpl; intros; [|destruct (list_mem)]; auto.
Qed.

Lemma list_disj_app_rD: forall l l1 l2,
  list_disj l (l1 ++ l2) -> list_disj l l1 /\ list_disj l l2.
Proof.
  induction l1; simpl; intros; [|destruct (list_mem)]; auto; tauto.
Qed.
Implicit Arguments list_disj_app_rD [l l1 l2].

Lemma list_disj_app_r: forall l l1 l2,
  list_disj l (l1 ++ l2) <-> list_disj l l1 /\ list_disj l l2.
Proof.
  induction l1; simpl; intros; [|destruct (list_mem)]; auto; tauto.
Qed.

Lemma list_disj_app_l: forall l1 l2 l,
  list_disj (l1 ++ l2) l <-> list_disj l1 l /\ list_disj l2 l.
Proof.
  split.
  - intros H; apply list_disj_comm, list_disj_app_r in H
    ; destruct H; auto using list_disj_comm.
  - intros [H1 H2]; apply list_disj_comm, list_disj_app_r
    ; auto using list_disj_comm.
Qed.

Lemma list_disj_minus_l: forall l l1 l2,
  list_disj l (list_minus l1 l2) -> list_disj (list_minus l l2) l1.
Proof.
  intros; apply list_disj_comm.
  revert l1 H; induction l as [|a l]; simpl in *; intros; auto.
  apply list_disj_cons_l in H; rewrite list_mem_minus in *.
  destruct (list_mem a l2); simpl; 
  destruct (list_mem a l1); simpl; intuition auto; discriminate.
Qed.

Lemma list_disj_minus_r: forall l l1 l2,
  list_disj (list_minus l1 l2) l -> list_disj l1 (list_minus l l2).
Proof.
  auto using list_disj_comm, list_disj_minus_l.
Qed.

Lemma list_disj_mon_l: forall l1 l2 l
    (DJ: list_disj l1 l)
    (IN: forall x, In x l2 -> In x l1),
  list_disj l2 l.
Proof.
  intros; induction l as [|a l IH]; simpl in *; auto. 
  specialize (IN a).
  pose proof (list_mem_spec a l1).
  pose proof (list_mem_spec a l2).
  do 2 destruct list_mem; auto; contradiction.
Qed.

Lemma list_disj_mon_r: forall l1 l2 l
    (DJ: list_disj l l1)
    (IN: forall x, In x l2 -> In x l1),
  list_disj l l2.
Proof.
  eauto using list_disj_mon_l, list_disj_comm.
Qed.

Lemma list_disj_spec: forall l1 l2,
  list_disj l1 l2 <-> (forall x, In x l1 -> ~ In x l2).
Proof.
  intros; revert l1; induction l2; simpl; intuition; subst.
  - pose proof (list_mem_spec x l1); destruct list_mem; firstorder. 
  - pose proof (list_mem_spec a l1); destruct list_mem; firstorder.
  - pose proof (list_mem_spec a l1); destruct list_mem; firstorder.
Qed.

Lemma list_assoc_flip: forall l' l x
    (DJ: list_disj (map (@fst _ _) l) (map (@fst _ _) l')),
  list_assoc x (l' ++ l) = list_assoc x (l ++ l').
Proof.
  intros; rewrite !list_assoc_app.
  pose proof (proj1 (list_disj_spec _ _) DJ x). 
  pose proof (list_assoc_spec x l).
  pose proof (list_assoc_spec x l').
  do 2 destruct list_assoc; auto.
  destruct H; apply in_map_fst; eauto. 
Qed.


(** *** Properties of [remove]. *)

Lemma remove_length: forall X eqd (x: X) xs,
  length (remove eqd x xs) <= length xs.
Proof.
  induction xs; s; auto.
  des_ifs; s; omega.
Qed.

Lemma remove_in: forall X eqd xs (x: X) y,
  In y (remove eqd x xs) <-> In y xs /\ ~ eqd y x.
Proof.
  induction xs; split; i; des; ss; vauto.
  - des_ifs.
    + hdes; eauto.
    + ss; des; subst; ss; des; subst; hdes; splits; eauto.
      destruct (eqd y x); vauto.
  - des; subst.
    + des_ifs; destruct (eqd _ _); vauto.
    + des_ifs.
      * apply IHxs; auto.
      * s; destruct (LEM (a = y)); auto.
        right; apply IHxs; auto.
Qed.


(** *** Properties of [list_max]. *)

Lemma list_max_in: forall ns n
    (IN: In n ns),
  n <= list_max ns.
Proof.
  induction ns; vauto; i; ss; des; subst; [by eauto with arith|].
  eapply le_trans; [eapply IHns|]; eauto with arith.
Qed.

Lemma list_max_remove: forall x xs,
  <<IN: In x xs -> list_max xs = max x (list_max (remove eqdec x xs))>> /\
  <<NIN: ~ In x xs -> list_max xs = list_max (remove eqdec x xs)>>.
Proof.
  induction xs; vauto; split; intro IN.
  - apply in_consD in IN; des; s; subst; des_ifs.
    + destruct (LEM (In a xs)); eauto.
      rewrite IN0, Max.max_assoc, Max.max_idempotent; auto.
    + s; rewrite Max.max_assoc, (Max.max_comm x), <-Max.max_assoc, IN0; auto.
  - des; s; des_ifs; ss; [exfalso|]; eauto.
Qed.

Lemma list_max_ext: forall l1 l2
    (EQ: forall x, In x l1 <-> In x l2),
  list_max l1 = list_max l2.
Proof.
  induction l1 as [? IH] using (well_founded_ind lt_length_well_founded); i.
  destruct l1.
  - destruct l2; [by vauto|]; exfalso; specialize (EQ n); des; ss; eauto.
  - dup EQ; specialize (EQ n); des.
    hexploit EQ; [by s; eauto|]; i; clear EQ EQ1.
    hexploit (list_max_remove n); i; des.
    rewrite IN; [|by s; eauto].
    rewrite (proj1 (list_max_remove _ _) X); f_equal.
    s; des_ifs; apply IH; [by s; apply le_n_S; auto using remove_length|].
    clear IN NIN; split; intro IN.
    + apply remove_in; apply remove_in in IN; des; split; auto.
      apply EQ0; s; eauto.
    + apply remove_in; apply remove_in in IN; des; split; auto.
      apply EQ0 in IN; ss; des; subst; auto.
      exfalso; destruct eq_nat_dec; vauto.
Qed.


(** *** Properties of [sublist]. *)

Lemma sublist_refl: forall X {M: EqDec X} (xs: list X),
  sublist xs xs.
Proof.
  induction xs; auto using sublist_nil, sublist_cons_keep.
Qed.
Hint Resolve sublist_refl.

Lemma sublist_nil': forall X {M: EqDec X} l,
  sublist nil l.
Proof.
  induction l; auto using sublist_nil, sublist_cons_drop.
Qed.

Lemma sublist_app X {M: EqDec X} : forall (la la' lb lb': list X)
    (SBa: sublist la la')
    (SBb: sublist lb lb'),
  sublist (la ++ lb) (la' ++ lb').
Proof.
  clear.
  i; ginduction SBa; vauto.
  - ginduction l; ss; i; eauto using sublist_cons_drop.
  - i; rewrite <-!app_comm_cons; eapply sublist_cons_keep; eauto.
Qed.

Lemma sublist_in: forall X {M: EqDec X} l l' x
    (SB: sublist l l')
    (IN: In x l),
  In x l'.
Proof.
  clear.
  i; ginduction SB; i; [by vauto|by right; auto|].
  ss; des; subst; auto.
Qed.

Lemma sublist_trans: forall X {M: EqDec X} (xs ys zs: list X)
    (SLx: sublist xs ys)
    (SLy: sublist ys zs),
  sublist xs zs.
Proof.
  clear.
  i; ginduction SLy; [by vauto| |]; i; inv SLx; eauto.
Qed.

Lemma sublist_r_app_l: forall A {M: EqDec A} xs xs' ys
    (SL: sublist xs xs'),
  sublist xs (xs' ++ ys).
Proof.
  i; lhs rewrite <-(app_nil_r xs); eauto using sublist_app, sublist_nil'.
Qed.

Lemma sublist_r_app_r: forall A {M: EqDec A} xs ys ys'
    (SL: sublist ys ys'),
  sublist ys (xs ++ ys').
Proof.
  i; lhs rewrite <-(app_nil_l ys); eauto using sublist_app, sublist_nil'.
Qed.






















(** *** Properties of [list_index]. *)

Lemma list_index_some: forall X {M: EqDec X} (x: X) xs
    (IN: In x xs),
  exists n, list_index x xs = Some n.
Proof.
  i; induction xs; vauto.
  apply in_consD in IN; des; subst; s; des_if; vauto.
  hexploit IHxs; auto; i; des; rewrite X0; s; eauto.
Qed.

Lemma list_index_length: forall X {M: EqDec X} x xs n
    (IN: list_index x xs = Some n),
  n < length xs.
Proof.
  induction xs; s; vauto; i.
  des_ifH IN; ss; clarify; auto with arith.
  destruct (list_index _ _); ss; clarify; eauto using lt_n_S.
Qed.


(** *** Properties of [firstn]. *)

Lemma firstn_nil: forall X n m (l: list X)
    (FI: firstn m l = nil)
    (LE: n <= m),
  firstn n l = nil.
Proof.
  clear.
  i; ginduction l; i.
  - destruct n; auto.
  - destruct LE; s; eauto.
    destruct m; eauto; vauto.
Qed.

Lemma firstn_firfstn: forall X n m (l: list X)
    (LE: n <= m),
  firstn n (firstn m l) = firstn n l.
Proof.
  induction n; auto; []; i.
  destruct (firstn m l) eqn:E.
  - (lhs s); erewrite firstn_nil; eauto; omega.
  - rewrite <-E.
    destruct m; [by vauto|]; s.
    destruct l; [by vauto|]; s in E.
    rewrite IHn; eauto; omega.
Qed.

Lemma firstn_S: forall X n l (d: X)
    (LEN: n < length l),
  firstn (S n) l = firstn n l ++ [hd d (skipn n l)].
Proof.
  clear.
  i; erewrite <-(firstn_skipn n (firstn _ _)).
  rewrite firstn_firfstn; [|omega].
  f_equal; ginduction n; i.
  - destruct l; ss; auto; omega.
  - destruct l; ss; [omega|].
    eapply lt_S_n, IHn in LEN; eauto.
Qed.


(** *** Properties of [skipn]. *)

Lemma skipn_le: forall X n (xs: list X)
    (LE: n < length xs),
  exists x, skipn n xs = x :: skipn (S n) xs.
Proof.
  clear.
  i; ginduction xs; s; [i; exfalso; omega|].
  destruct n; [by vauto|]; i.
  apply lt_S_n in LE; hexploit IHxs; eauto.
Qed.

Lemma skipn_index: forall X {M: EqDec X} n x xs
    (IN: list_index x xs = Some n),
  exists xs', skipn n xs = x :: xs'.
Proof.
  clear.
  i; ginduction xs; [by auto|]; i.
  ss; des_ifH IN; clarify; s; [by eauto|].
  destruct (list_index _ _) eqn:E; [|by vauto]; ss; clarify; s.
  hexploit IHxs; eauto.
Qed.

Lemma skipn_list_assoc: forall x l y n
    (IN: list_assoc x l = Some y)
    (IDX: list_index x (map fst l) = Some n),
  exists l', skipn n l = (x, y) :: l'.
Proof.
  clear; i.
  hexploit (skipn_index _ _ IDX); i; des.
  ginduction l; destruct n; vauto
  ; destruct a; i; ss; clarify; ss; des_ifH IN; vauto; des_ifH IDX; vauto.
  hexploit IHl; eauto; destruct (list_index _ _); ss; vauto.
Qed.

Lemma skipn_cons: forall x l y n d
    (IN: list_assoc x l = Some y)
    (LE: S n <= withd d (list_index x (map fst l))),
  exists x' y', skipn n l = (x', y') :: skipn (S n) l.
Proof.
  i; assert (list_index x (map fst l) = Some (withd d (list_index x (map fst l)))).
  { apply list_assoc_someD, (in_map fst) in IN.
    clear -IN; induction (map fst l); vauto.
    ss; des; subst; des_if; vauto; rewrite IHl0; eauto.
  }
  hexploit skipn_le.
  - eapply lt_trans; eauto.
    instantiate (1 := l).
    rewrite <-(map_length fst); eauto using list_index_length.
  - i; des; destruct x0; eauto.
Qed.

Lemma skipn_length: forall X (xs: list X) n,
  length (skipn n xs) = length xs - n.
Proof.
  clear.
  i; ginduction n; i; s; eauto with arith.
  destruct xs; s; vauto.
Qed.

Lemma skipn_eq_nil: forall X n (l: list X)
    (LEN: length l <= n),
  skipn n l = nil.
Proof.
  induction n; destruct l; auto; s; i; [|apply IHn]; omega.
Qed.

Lemma skipn_app: forall X (l l': list X) n,
  skipn n (l ++ l') = skipn n l ++ skipn (n - length l) l'.
Proof.
  induction l; destruct n; s; eauto.
Qed.

Lemma skipn_nil: forall X n m (l: list X)
    (SK: skipn n l = nil)
    (LE: n <= m),
  skipn m l = nil.
Proof.
  clear.
  i; ginduction l; i.
  - destruct m; auto.
  - destruct LE; s; eauto.
    destruct n; eauto; vauto.
Qed.

Lemma skipn_tl: forall X (l: list X) n,
  skipn (S n) l = tl (skipn n l).
Proof.
  induction l; i.
  - s; erewrite skipn_nil; [|insts $ 0|]; eauto; omega.
  - destruct n; s; auto.
    rewrite <-IHl; s; auto.
Qed.

Lemma skipn_skipn: forall X n m (l: list X),
  skipn n (skipn m l) = skipn (n + m) l.
Proof.
  induction n; auto; []; i.
  destruct (skipn m l) eqn:E.
  - (lhs s); erewrite skipn_nil; eauto; omega.
  - assert (SK := skipn_tl l m).
    rewrite E in SK; simpl tl in SK.
    simpl skipn; rewrite <-SK, IHn.
    rewrite plus_comm; s; rewrite plus_comm; auto.
Qed.


(** *** Properties of [repeat]. *)

Lemma repeat_S: forall a n,
  repeat a (S n) = repeat a n ++ [a].
Proof.
  induction n; auto.
  i; s; rewrite <-IHn; auto.
Qed.

Lemma repeat_split: forall a n m
    (LE: n <= m),
  repeat a m = repeat a n ++ repeat a (m-n).
Proof.
  i; induction LE.
  - rewrite <-minus_n_n; s; rewrite app_nil_r; auto.
  - rewrite repeat_S, IHLE, <-app_assoc; f_equal.
    replace (S m - n) with (S (m - n)); [by rewrite repeat_S; auto|].
    induction n; auto with arith.
Qed.

Lemma repeat_length: forall a n,
  length (repeat a n) = n.
Proof.
  induction n; s; auto.
Qed.


(** *** Properties of [combine_full]. *)

Lemma combine_full_in: forall X Y (xs: list X) (ys: list Y) x y
    (INx: In x xs)
    (INy: In y ys),
  In (x, y) (combine_full xs ys).
Proof.
  i; induction xs; auto.
  s; s in INx; des; subst; apply in_or_app; [left|by right; eauto].
  clear IHxs; induction ys; auto.
  s in INy; des; subst; s; [|right; eapply IHys]; eauto.
Qed.


(** *** Lemmas for automation. *)

Lemma list_disj_map_fst_minus2: forall l l' l'',
  list_disj (list_minus (map (@fst A B) l) l') l'' ->
  list_disj (map (@fst A B) (list_minus2 l l')) l''.
Proof.
  intros; rewrite list_map_fst_minus2; trivial.
Qed.

Lemma list_disj_minus2_mon: forall l l' l'',
  list_disj (map (@fst A B) l) (map (@fst A B) l') -> 
  list_disj (map (@fst A B) (list_minus2 l l''))
            (map (@fst A B) (list_minus2 l' l'')).
Proof.
  intros; eapply list_disj_mon_l; [eapply (list_disj_mon_r _ _ _ H)|];
  intros x X; apply in_map_fst in X; destruct X as [y X];
  apply in_map_fst; exists y; eapply in_list_minus2 in X; tauto.
Qed.

End ListDefs.

Hint Resolve list_assoc_cons_eq.
Hint Resolve list_assoc_cons_neq.
Hint Constructors sublist.
Hint Resolve sublist_refl.
Hint Resolve sublist_r_app_l.
Hint Resolve sublist_r_app_r.

Create HintDb subst_lemmas.
Hint Resolve list_disj_minus_l: subst_lemmas.
Hint Resolve list_disj_minus2_mon: subst_lemmas.
Hint Resolve list_disj_map_fst_minus2: subst_lemmas.
Hint Rewrite list_assoc_remdup2 list_minus2_remdup2: subst_rw_lemmas. 


(** ** More properties of list operations. *)

Lemma list_assoc_eqdom: forall A deq X Y
    (xs: list (A * X)) (ys: list (A * Y))
    (DOM: map fst xs = map fst ys)
    a x (LU: list_assoc deq a xs = Some x),
  exists y, list_assoc deq a ys = Some y.
Proof.
  i; apply list_assoc_map; rewrite <-DOM.
  apply in_map_iff; apply list_assoc_someD in LU.
  esplits; eauto; s; eauto.
Qed.

Lemma list_assoc_eqdom_false: forall A deq X Y
    (xs: list (A * X)) ys a (y: Y)
    (DOM: map fst xs = map fst ys)
    (LUx: list_assoc deq a xs = None)
    (LUy: list_assoc deq a ys = Some y),
  False.
Proof.
  i; eapply list_assoc_noneD; eauto.
  rewrite DOM; fold (fst (a, y)); apply (in_map fst).
  eauto using list_assoc_someD.
Qed.

Lemma list_assoc_map_pair: forall X X' {M: EqDec X} {M': EqDec X'} Y
    (f: X -> X') (x: X) (y: Y) (l: list (X * Y)) x'
    (IN: list_assoc eqdec x l = Some y)
    (INJ: forall xa xb (EQ: f xa = f xb), xa = xb)
    (E: x' = f x),
  list_assoc eqdec x' (map_pair f id l) = Some y.
Proof.
  induction l; auto; i.
  destruct a; s; subst; des_ifs.
  - apply INJ in e; subst; rewrite list_assoc_cons_eq in IN; clarify.
  - rewrite list_assoc_cons_neq in IN; eauto; ii; subst; eauto.
Qed.

Lemma list_assoc_map_pairD: forall X {M: EqDec X} Y Z (g: Z -> Y) l (x: X) y
    (IN: list_assoc eqdec x (map_pair id g l) = Some y),
  exists z, list_assoc eqdec x l = Some z /\ y = g z.
Proof.
  i; ginduction l; vauto.
  i; s in IN; destruct a; des_ifH IN; clarify; eauto.
  rewrite list_assoc_cons_neq; auto.
Qed.

Lemma repeat_map: forall X Y (a: X) n (f: X -> Y),
  map f (repeat a n) = repeat (f a) n.
Proof.
  i; induction n; s; try f_equal; auto.
Qed.


(** ** Tactics. *)

Ltac subst_tac :=
  solve [simpl; intros; repeat match goal with 
   | _ => solve [trivial | contradiction | discriminate] 
   | H : list_disj _ _ (app _ _) |- _ => apply list_disj_app_r in H; destruct H
   | _ => progress f_equal
   | H : context[list_mem _ _ (map (@fst _ _) _)] |- _ => rewrite list_mem_map_fst in H
   | |- context[match ?x with Some _ => _ | None => _ end] => destruct x
   | |- context[if ?x then _ else _] => destruct x
  end; auto with subst_lemmas].

Ltac subst_tac' :=
  try solve [simpl; intros; 
  repeat (match goal with 
          | _ => solve [congruence | eauto with subst_lemmas]
          | _ => progress autorewrite with subst_rw_lemmas
          | |- context[if ?x then _ else _] => 
            destruct x; simpl in *
          | _ => progress f_equal
          end)].

Ltac sng :=
  repeat match goal with
    | H: list_assoc _ ?x [?p] = Some ?y |- _
      => apply list_assoc_sng in H; desH H
    | H: Some ?y = list_assoc _ ?x [?p] |- _
      => rewrite eq_sym in H
  end; subst; clean.



(** * Additions to the Arith library. *)


(** ** Properties of [minus]. *)

Lemma plus_le_minus: forall m p n
    (LE: m + p <= n),
  m <= n - p.
Proof.
  i; omega.
Qed.

Lemma plus_minus_eq: forall m n p
    (LE: p <= m),
  m + n - p = m - p + n.
Proof.
  i; omega.
Qed.

Lemma minus_minus: forall m n p,
  m - (n + p) = m - n - p.
Proof.
  i; omega.
Qed.


(** ** Alternative definition of the standard [le] relation (less-than-or-equal).

    Useful for doing induction on. *)

Inductive myle (n: nat) : nat -> Prop :=
| myle_n : myle n n
| myle_S : forall m, myle (S n) m -> myle n m
.
Hint Constructors myle.

Lemma myle_n_S: forall m n
    (LE: myle m n),
  myle (S m) (S n).
Proof.
  i; induction LE; auto with arith.
Qed.

Lemma myle_le: forall m n,
  myle m n <-> m <= n.
Proof.
  split; i; induction H; auto using myle_n_S with arith.
Qed.



(** * Option type extended with error case. *)


(** ** Definition. *)

Inductive eoption X :=
| Just (x: X)
| Nothing
| Error
.
Hint Constructors eoption.
Arguments Error [X].
Arguments Nothing [X].

Definition eoplus X (eo eo': eoption X): eoption X :=
  match eo with
    | Just x =>
      match eo' with
        | Nothing => eo
        | _ => Error
      end
    | Nothing => eo'
    | Error => eo
  end.


(** ** Properties.

    [eoption X] forms a commutative monoid.
*)

Lemma eoplus_nothing: forall X (eo: eoption X),
  eoplus eo Nothing = eo.
Proof.
  i; destruct eo; eauto.
Qed.

Lemma eoplus_comm: forall X (eo eo': eoption X),
  eoplus eo eo' = eoplus eo' eo.
Proof.
  i; destruct eo, eo'; eauto.
Qed.

Lemma eoplus_assoc: forall X (eo eo' eo'': eoption X),
  eoplus eo (eoplus eo' eo'') = eoplus (eoplus eo eo') eo''.
Proof.
  i; destruct eo, eo', eo''; eauto.
Qed.

Ltac nonothing :=
  repeat match goal with
    | H: context[eoplus ?X Nothing] |- _ => rewrite (eoplus_nothing X) in H
    | |- context[eoplus ?X Nothing]      => rewrite (eoplus_nothing X)
    | H: context[eoplus Nothing ?X] |- _ => rewrite (eoplus_comm Nothing X) in H
    | |- context[eoplus ?X Nothing]      => rewrite (eoplus_comm Nothing X)
  end.



(** * Additions to the Relations library. *)

Notation rtc := (clos_refl_trans _). 
Notation tc := (clos_trans _). 
Hint Immediate rt_step rt_refl t_step.

Lemma tc_well_founded: forall A R
    (WF: @well_founded A R),
  well_founded (clos_trans _ R).
Proof. clear.
  ii; induction a using (well_founded_ind WF); econs; i.
  apply clos_trans_tn1_iff in H0; induction H0; [by auto|].
  apply IHclos_trans_n1; specialize (H y0 H0); invc H; eauto.
Qed.

Lemma rtc_mon: forall X (R R': X -> X -> Prop) a b
    (IN: rtc R a b)
    (LE: R <2= R'),
  rtc R' a b.
Proof.
  i; induction IN; [econs 1|econs 2|econs 3]; eauto.
Qed.

Lemma tc_mon: forall X (R R': X -> X -> Prop) a b
    (IN: tc R a b)
    (LE: R <2= R'),
  tc R' a b.
Proof.
  i; induction IN; [econs 1|econs 2]; eauto.
Qed.

Lemma tc_rtc: forall X r (a b: X)
    (TC: clos_trans _ r a b),
  clos_refl_trans _ r a b.
Proof.
  i; induction TC; eauto using rt_trans.
Qed.

Lemma rtc_tc: forall X r (a b: X)
    (RTC: rtc r a b),
  a = b \/ tc r a b.
Proof.
  i; induction RTC; des; subst; eauto.
  right; econs 2; eauto.
Qed.

Lemma tc_rtc_tc: forall X r (a b c: X)
    (TC: tc r a b)
    (RTC: rtc r b c),
  tc r a c.
Proof.
  i; induction RTC; eauto using t_trans, t_step.
Qed.

Lemma rtc_tc_tc: forall X r (a b c: X)
    (RTC: rtc r a b)
    (TC: tc r b c),
  tc r a c.
Proof.
  i; induction RTC; eauto using t_trans, t_step.
Qed.

Lemma tc_first: forall X r (a c: X)
    (CT: tc r a c),
  exists b, r a b /\ rtc r b c.
Proof.
  i; ginduction CT; eauto using rt_refl; clear IHCT2; des.
  eauto using rt_step, tc_rtc, rtc_tc_tc.
Qed.

Lemma tc_last: forall X r (a c: X)
    (CT: tc r a c),
  exists b, rtc r a b /\ r b c.
Proof.
  i; induction CT; eauto using rt_refl; clear IHCT1; des.
  eauto using rt_step, tc_rtc, tc_rtc_tc.
Qed.
