(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** ML-like source language *)

Require Import Arith Bool List Program Omega.
Require Import vgtac basis lang_common types.

Set Implicit Arguments.



(** * Definitions. *)



(** ** Syntax.

    Note that there are no typing annotations.
*)

Definition location := nat.

Inductive exp : Set :=
| Einput
| Eoutput (e:exp)
| Evar (x:termvar)
| Elbl (l:label)
| Eunit
| Enat (n:nat)
| Ebop (op: binop) (ea: exp) (eb: exp)
| Eifnz (e:exp) (ea:exp) (eb:exp)
| Epair (ea:exp) (eb:exp)
| Efst (e:exp)
| Esnd (e:exp)
| Einl (e:exp)
| Einr (e:exp)
| Ecase (e:exp) (x:termvar) (ea:exp) (eb:exp)
| Eroll (e:exp)
| Eunroll (e:exp)
| Efix (f:termvar) (x:termvar) (e:exp)
| Eapp (ea:exp) (eb:exp)
| Egen (e:exp)
| Einst (e:exp)
| Epack (e:exp)
| Eunpack (x:termvar) (ea:exp) (eb:exp)
| Eloc (loc:location)
| Enew (e:exp)
| Eread (e:exp)
| Ewrite (ea:exp) (eb:exp)
| Erefeq (ea:exp) (eb:exp)
.


(** *** Free variables. *)

Fixpoint fv_exp (ee:exp) (xx: termvar) : Prop :=
  match ee with
  | Eoutput e => fv_exp e xx
  | Evar x => xx = x
  | Ebop op ea eb => fv_exp ea xx \/ fv_exp eb xx
  | Eifnz e ea eb => fv_exp e xx \/ fv_exp ea xx \/ fv_exp eb xx
  | Epair ea eb => fv_exp ea xx \/ fv_exp eb xx
  | Efst e => fv_exp e xx
  | Esnd e => fv_exp e xx
  | Einl e => fv_exp e xx
  | Einr e => fv_exp e xx
  | Ecase e x ea eb => fv_exp e xx \/ 
      ((~ xx = x) /\ (fv_exp ea xx \/ fv_exp eb xx))
  | Eroll e => fv_exp e xx
  | Eunroll e => fv_exp e xx
  | Efix f x e => (~ xx = f) /\ (~ xx = x) /\ fv_exp e xx
  | Eapp ea eb => fv_exp ea xx \/ fv_exp eb xx
  | Egen e => fv_exp e xx
  | Einst e => fv_exp e xx
  | Epack e => fv_exp e xx
  | Eunpack x ea eb => (fv_exp ea xx) \/ ((~ xx = x) /\ fv_exp eb xx)
  | Eloc loc => False
  | Enew e => ((fv_exp e xx))
  | Eread e => ((fv_exp e xx))
  | Ewrite ea eb => (fv_exp ea xx) \/ (fv_exp eb xx)
  | Erefeq ea eb => (fv_exp ea xx) \/ (fv_exp eb xx)
  | _ => False
end.


(** *** Substitutions.

    Modelled simply as total functions from variables to expressions.
*)

Definition subst_closed (sub: termvar -> exp) :=
  forall x a, ~ fv_exp (sub x) a.
Hint Unfold subst_closed.

Definition subst_weakly_closed (sub: termvar -> exp) :=
  forall x a (FV: fv_exp (sub x) a), a = x.
Hint Unfold subst_weakly_closed.

(** Apply substitution to expression. *)
Fixpoint subst_exp (sub: termvar -> exp) (ee:exp) {struct ee} : exp :=
  match ee with
  | Eoutput e => Eoutput (subst_exp sub e)
  | Evar x => sub x
  | Ebop op ea eb => Ebop op (subst_exp sub ea) (subst_exp sub eb)
  | Eifnz e ea eb =>
    Eifnz (subst_exp sub e) (subst_exp sub ea) (subst_exp sub eb)
  | Epair ea eb => Epair (subst_exp sub ea) (subst_exp sub eb)
  | Efst e => Efst (subst_exp sub e)
  | Esnd e => Esnd (subst_exp sub e)
  | Einl e => Einl (subst_exp sub e)
  | Einr e => Einr (subst_exp sub e)
  | Ecase e x ea eb =>
    Ecase (subst_exp sub e) x (subst_exp [Evar x//x]>>sub ea)
          (subst_exp [Evar x//x]>>sub eb)
  | Eroll e => Eroll (subst_exp sub e)
  | Eunroll e => Eunroll (subst_exp sub e)
  | Efix f x e => Efix f x (subst_exp [Evar x//x]>>[Evar f//f]>>sub e)
  | Eapp ea eb => Eapp (subst_exp sub ea) (subst_exp sub eb)
  | Egen e => Egen (subst_exp sub e)
  | Einst e => Einst (subst_exp sub e)
  | Epack e => Epack (subst_exp sub e)
  | Eunpack x ea eb => Eunpack x (subst_exp sub ea) 
      (subst_exp [Evar x//x]>>sub eb)
  | Enew e => Enew (subst_exp sub e)
  | Eread e => Eread (subst_exp sub e)
  | Ewrite ea eb => Ewrite (subst_exp sub ea) (subst_exp sub eb)
  | Erefeq ea eb => Erefeq (subst_exp sub ea) (subst_exp sub eb)
  | e => e
end.



(** ** Operational semantics.

    Fairly standard substitution-based semantics.  External references
    (labels) are resolved by lookup in an environment, which counts as
    a step of computation.
*)


(** *** Run-time values. *)

Fixpoint is_value (ee:exp) : Prop :=
  match ee with
  | Evar x => True
  | Eunit => True
  | Enat n => True
  | Epair ea eb => is_value ea /\ is_value eb
  | Einl e => is_value e
  | Einr e => is_value e
  | Eroll e => is_value e
  | Efix f x e => True
  | Egen e => True
  | Epack e => is_value e
  | Eloc loc => True
  | _ => False
end.

Structure val :=
  mkval { e_of :> exp; val_is_value : is_value e_of }.
Hint Resolve val_is_value.

Program Canonical Structure Vvar x := 
  @mkval (Evar x) _.
Program Canonical Structure Vunit := 
  @mkval (Eunit) _.
Program Canonical Structure Vnat n := 
  @mkval (Enat n) _.
Program Canonical Structure Vpair (va vb: val) := 
  @mkval (Epair va vb) _.
Program Canonical Structure Vinl (va: val) :=
  @mkval (Einl va) _.
Program Canonical Structure Vinr (vb: val) :=
  @mkval (Einr vb) _.
Program Canonical Structure Vroll (v: val) :=
  @mkval (Eroll v) _.
Program Canonical Structure Vfix f x e := 
  @mkval (Efix f x e) _.
Program Canonical Structure Vgen e := 
  @mkval (Egen e) _.
Program Canonical Structure Vpack (v: val) :=
  @mkval (Epack v) _.
Program Canonical Structure Vloc l := 
  @mkval (Eloc l) _.

Definition heap := @pmap location _ val.
Definition env := list (label * val).
Definition vid (x: termvar) := Vvar x. 


(** *** Syntactic sugar. *)

Definition Elet x e1 e2 :=
  Eapp (Vfix x x e2) e1.

Definition Vabs x e :=
  Vfix x x e.


(** *** Evaluation contexts ("continuations"). *)

Inductive scont : Set :=
| SKoutput
| SKbopL (op: binop) (e:exp)
| SKbopR (op: binop) (v:val)
| SKifnz (ea:exp) (eb:exp)
| SKpairL (e:exp)
| SKpairR (v:val)
| SKfst
| SKsnd
| SKinl
| SKinr
| SKcase (x:termvar) (ea:exp) (eb:exp)
| SKroll
| SKunroll
| SKappL (e:exp)
| SKappR (v:val)
| SKinst
| SKpack
| SKunpack (x:termvar) (e:exp)
| SKnew
| SKread
| SKwriteL (e:exp)
| SKwriteR (v:val)
| SKrefeqL (e:exp)
| SKrefeqR (v:val)
.

Definition cont : Set := list scont.

Definition app_scont (SK:scont) (ee:exp) : exp :=
  match SK with
  | SKoutput => Eoutput ee
  | SKbopL op e => Ebop op ee e
  | SKbopR op v => Ebop op v ee
  | SKifnz ea eb => Eifnz ee ea eb
  | SKpairL e => Epair ee e
  | SKpairR v => Epair v ee
  | SKfst => Efst ee
  | SKsnd => Esnd ee
  | SKinl => Einl ee
  | SKinr => Einr ee
  | SKcase x ea eb => Ecase ee x ea eb
  | SKroll => Eroll ee
  | SKunroll => Eunroll ee
  | SKappL e => Eapp ee e
  | SKappR v => Eapp v ee
  | SKinst => Einst ee
  | SKpack => Epack ee
  | SKunpack x e => Eunpack x ee e
  | SKnew => Enew ee
  | SKread => Eread ee
  | SKwriteL e => Ewrite ee e
  | SKwriteR v => Ewrite v  ee
  | SKrefeqL e => Erefeq ee e
  | SKrefeqR v => Erefeq v ee
end.

Definition app_cont (K : cont) (e : exp) : exp :=
  fold_left (fun e SK => app_scont SK e) K e.

Definition app_cont_r K e := fold_right app_scont e K.











(** *** Small-step reduction. *)

Inductive redok : event -> heap -> exp -> heap -> exp -> Prop :=
| Rinput : forall h (n: nat),
   redok (ein n) h Einput h (Enat n)
| Routput : forall h (n: nat),
   redok (eout n) h (Eoutput (Enat n)) h Eunit
| Rbop : forall h op (na nb: nat),
   redok etau h (Ebop op (Enat na) (Enat nb)) h (Enat (bosem op na nb))
| Rifz : forall h (ea eb:exp),
  redok etau h (Eifnz (Enat 0) ea eb) h eb
| Rifnz : forall h (ea eb:exp) (n: nat),
  redok etau h (Eifnz (Enat (S n)) ea eb) h ea
| Rfst : forall h (va vb:val),
  redok etau h (Efst (Epair va vb)) h va
| Rsnd : forall h (va vb:val),
  redok etau h (Esnd (Epair va vb)) h vb
| Rinl : forall h (x:termvar) (v:val) (ea eb:exp),
  redok etau h (Ecase (Einl v) x ea eb) h (subst_exp [e_of v//x]>>vid ea) 
| Rinr : forall h (x:termvar) (v:val) (ea eb:exp),
  redok etau h (Ecase (Einr v) x ea eb) h (subst_exp [e_of v//x]>>vid eb) 
| Runroll : forall h (v:val),
  redok etau h (Eunroll ((Eroll v))) h v 
| Rbeta : forall h (f x:termvar) (e:exp) (v:val),
  redok etau h (Eapp ((Efix f x e)) v) 
    h (subst_exp([e_of v//x]>>[Efix f x e//f]>>vid) e)
| Rtybeta : forall h (e:exp),
  redok etau h (Einst ((Egen e))) h e
| Runpack : forall h (x:termvar) (v:val) (e:exp),
  redok etau h (Eunpack x (Epack v) e) h (subst_exp [e_of v//x]>>vid e)
| Rnew : forall h (loc:location) (v:val)
    (PV: pmvalid ({loc |-> v} ## h)),
  redok etau h (Enew v) ({loc |-> v} ## h) (Eloc loc)
| Rread : forall h (loc:location) (v:val)
    (RD: pmread h loc v),
  redok etau h (Eread (Eloc loc)) h v
| Rwrite : forall h (loc:location) (v v':val),
  redok etau ({loc |-> v} ## h) (Ewrite (Eloc loc) v') 
    ({loc |-> v'} ## h) Eunit
| Rrefeq : forall h (loca locb:location),
  redok etau h (Erefeq (Eloc loca) (Eloc locb))
    h (Enat (if eqdec loca locb then 1 else 0))
.
Hint Constructors redok. 

Inductive reduce_ok : event -> env -> heap -> exp -> heap -> exp -> Prop := 
| Rbase: forall evt ha ea hb eb env
      (RED: redok evt ha ea hb eb : Prop),
    reduce_ok evt env ha ea hb eb
| Rlbl: forall h env l e (RD: lookup l env = Some e),
   reduce_ok etau env h (Elbl l) h e
| Rsctx: forall evt SK ha ea hb eb env
      (RED: reduce_ok evt env ha ea hb eb),
    reduce_ok evt env ha (app_scont SK ea) hb (app_scont SK eb) 
.
Hint Constructors reduce_ok.

Inductive src_step : event -> (heap * env * exp) -> (heap * env * exp) -> Prop :=
| Rok: forall evt env h e h' e'
      (RED: reduce_ok evt env h e h' e'),
    src_step evt (h, env, e) (h', env, e')
| Rerr: forall env h e
      (NVAL: ~ is_value e)
      (NRED: forall evt h' e', ~ reduce_ok evt env h e h' e'),
    src_step etau (h, env, e) (None, env, e)
.
Hint Constructors src_step.


(** *** Module loading. *)


Definition module := list (label * val).

Definition src_load (m: module) : option (heap * env * exp) :=
  match lookup MAIN m with
    | Some v => Some (pmemp, m, Eapp v Eunit)
    | None => None
  end.


(** ** Typing. *)

Inductive hastypepre (F: lblenv) : termenv -> exp -> typ -> Prop :=
| HTinput : forall G,
              hastypepre F G Einput Tnat
| HToutput : forall G e,
               hastypepre F G e Tnat ->
               hastypepre F G (Eoutput e) Tunit
| HTvar : forall (G:termenv) (x:termvar) (ty:typ),
            list_assoc eqdec x G = Some ty ->
            hastypepre F G (Evar x) ty
| HTlbl : forall (G:termenv) (l:label) (ty:typ),
            list_assoc eqdec l F = Some ty ->
            hastypepre F G (Elbl l) ty
| HTunit : forall (G:termenv),
             hastypepre F G Eunit Tunit
| HTint : forall (G:termenv) (n:nat),
            hastypepre F G (Enat n) Tnat
| HTbop : forall G op ea eb,
            hastypepre F G ea Tnat ->
            hastypepre F G eb Tnat ->
            hastypepre F G (Ebop op ea eb) Tnat
| HTifnz : forall (G:termenv) (e ea eb:exp) (ty:typ),
             hastypepre F G e Tnat ->
             hastypepre F G ea ty ->
             hastypepre F G eb ty ->
             hastypepre F G (Eifnz e ea eb) ty
| HTpair : forall (G:termenv) (ea eb:exp) (tya tyb:typ),
             hastypepre F G ea tya ->
             hastypepre F G eb tyb ->
             hastypepre F G (Epair ea eb) (Tprod tya tyb)
| HTfst : forall (G:termenv) (e:exp) (tya tyb:typ),
            hastypepre F G e (Tprod tya tyb) ->
            lc_typ 0 tyb ->
            hastypepre F G (Efst e) tya
| HTsnd : forall (G:termenv) (e:exp) (tyb tya:typ),
            hastypepre F G e (Tprod tya tyb) ->
            lc_typ 0 tya ->
            hastypepre F G (Esnd e) tyb
| HTinl : forall (G:termenv) (e:exp) (tya tyb:typ),
            hastypepre F G e tya ->
            hastypepre F G (Einl e) (Tsum tya tyb)
| HTinr : forall (G:termenv) (e:exp) (tya tyb:typ),
            hastypepre F G e tyb ->
            hastypepre F G (Einr e) (Tsum tya tyb)
| HTcase : forall (G:termenv) (x:termvar) (e ea eb:exp) (ty tya tyb:typ),
             hastypepre F G e (Tsum tya tyb) ->
             hastypepre F ((x, tya) :: G) ea ty ->
             hastypepre F ((x, tyb) :: G) eb ty ->
             lc_typ 0 tya ->
             lc_typ 0 tyb ->
             hastypepre F G (Ecase e x ea eb) ty
| HTroll : forall (G:termenv) (e:exp) (ty:typ),
             hastypepre F G e (tsubst_bvar 0 (Tmu ty) ty) ->
             hastypepre F G (Eroll e) (Tmu ty)
| HTunroll : forall (G:termenv) (e:exp) (ty:typ),
               hastypepre F G e (Tmu ty) ->
               hastypepre F G (Eunroll e) (tsubst_bvar 0 (Tmu ty) ty)
| HTfix : forall (G:termenv) (f x:termvar) (e:exp) (tya tyb:typ),
            hastypepre F ((x, tya) :: (f, Tarrow tya tyb) :: G) e tyb ->
            hastypepre F G (Efix f x e) (Tarrow tya tyb)
| HTapp : forall (G:termenv) (ea eb:exp) (tya tyb:typ),
            hastypepre F G ea (Tarrow tyb tya) ->
            hastypepre F G eb tyb ->
            lc_typ 0 tyb ->     
            hastypepre F G (Eapp ea eb) tya
| HTgen : forall (G:termenv) (e:exp) (ty:typ) (aa:typename),
            hastypepre F G e (tsubst_bvar 0 (Tname aa) ty) ->
            (~ ftn_typ_l (ty :: map (@snd _ _) F) aa) ->
            (~ ftn_typ_l (ty :: map (@snd _ _) G) aa) ->
            hastypepre F G (Egen e) (Tforall ty)
| HTinst : forall (G:termenv) (e:exp) (tya tyb:typ),
             hastypepre F G e (Tforall tya) ->
             lc_typ 0 tyb ->
             hastypepre F G (Einst e) (tsubst_bvar 0 tyb tya) 
| HTpack : forall (G:termenv) (e:exp) (tya tyb:typ),
             hastypepre F G e (tsubst_bvar 0 tyb tya) ->
             lc_typ 0 tyb ->
             hastypepre F G (Epack e) (Texists tya)
| HTunpack : forall (G:termenv) (x:termvar) (ea eb:exp) (tya tyb:typ) aa,
               hastypepre F G ea (Texists tya) ->
               lc_typ 1 tya ->
               (~ ftn_typ_l (tya :: tyb :: map (@snd _ _) F) aa) ->
               (~ ftn_typ_l (tya :: tyb :: map (@snd _ _) G) aa) ->
               hastypepre F ((x, tsubst_bvar 0 (Tname aa) tya) :: G) eb tyb ->
               hastypepre F G (Eunpack x ea eb) tyb
| HTnew : forall (G:termenv) (e:exp) (ty:typ),
            hastypepre F G e ty ->
            hastypepre F G (Enew e) (Tref ty)
| HTread : forall (G:termenv) (e:exp) (ty:typ),
             hastypepre F G e (Tref ty) ->
             hastypepre F G (Eread e) ty
| HTwrite : forall (G:termenv) (ea eb:exp) (ty:typ),
              hastypepre F G ea (Tref ty) ->
              hastypepre F G eb ty ->
              lc_typ 0 ty ->
              hastypepre F G (Ewrite ea eb) Tunit
| HTrefeq : forall (G:termenv) (ea eb:exp) (ty:typ),
              hastypepre F G ea (Tref ty) ->
              hastypepre F G eb (Tref ty) ->
              lc_typ 0 ty ->
              hastypepre F G (Erefeq ea eb) Tnat
              
.
Hint Constructors hastypepre.

Inductive hastype F G e ty : Prop :=
_hastype
  (HT: hastypepre F G e ty)
  (LCE: lc_env 0 F)
  (LCE: lc_env 0 G)
  (LCT: lc_typ 0 ty).
Hint Constructors hastype.

Inductive mhastypepre: lblenv -> module -> lblenv -> Prop :=
| MHTnil: forall FI, 
    mhastypepre FI nil nil
| MHTcons: forall FI FO l ty (v: val) m
      (MHT: mhastypepre (FI ++ [(l,ty)]) m FO)
      (WTV: hastype FI nil v ty)
      (NEW: lookup l FI = None),
    mhastypepre FI ((l,v)::m) ((l,ty)::FO)
.
Hint Constructors mhastypepre.

Inductive mhastype FI (m: module) FO : Prop :=
_mhastype
  (MHT: mhastypepre FI m FO)
  (FLI: is_thunkenv FI)
  (FLO: is_thunkenv FO)
.
Hint Constructors mhastype.



(** ** Canonical structure for KE split. TODO *)

Structure taggedExp := tExp0 { untagExp :> exp}.

Structure KEsplit := mkKEsplit 
  { term_of : taggedExp; scont_of : scont; body_of : exp; 
    KEsplit_pf : untagExp term_of = app_cont [scont_of] body_of }.

Definition tExp1  e := @tExp0  e.
Definition tExp2  e := @tExp1  e.
Definition tExp3  e := @tExp2  e.
Definition tExp4  e := @tExp3  e.
Definition tExp5  e := @tExp4  e.
Definition tExp6  e := @tExp5  e.
Definition tExp7  e := @tExp6  e.
Definition tExp8  e := @tExp7  e.
Definition tExp9  e := @tExp8  e.
Definition tExp10 e := @tExp9  e.
Definition tExp11 e := @tExp10 e.
Definition tExp12 e := @tExp11 e.
Definition tExp13 e := @tExp12 e.
Definition tExp14 e := @tExp13 e.
Definition tExp15 e := @tExp14 e.
Definition tExp16 e := @tExp15 e.
Definition tExp17 e := @tExp16 e.
Definition tExp18 e := @tExp17 e.
Definition tExp19 e := @tExp18 e.
Definition tExp20 e := @tExp19 e.
Definition tExp21 e := @tExp20 e.
Definition tExp22 e := @tExp21 e.
Canonical Structure tExp23 e := @tExp22 e.

Program Canonical Structure KES_SKoutput ee :=
  @mkKEsplit (tExp0 (Eoutput ee)) SKoutput ee _.
Program Canonical Structure KES_SKbopR op ee (ve: val) :=
  @mkKEsplit (tExp1 (Ebop op ve ee)) (SKbopR op ve) ee _.
Program Canonical Structure KES_SKbopL op ee e :=
  @mkKEsplit (tExp2 (Ebop op ee e)) (SKbopL op e) ee _.
Program Canonical Structure KES_SKifnz ee e1 e2 :=
  @mkKEsplit (tExp3 (Eifnz ee e1 e2)) (SKifnz e1 e2) ee _.
Program Canonical Structure KES_SKpairR ee (ve: val) :=
  @mkKEsplit (tExp4 (Epair ve ee)) (SKpairR ve) ee _.
Program Canonical Structure KES_SKpairL ee e :=
  @mkKEsplit (tExp5 (Epair ee e)) (SKpairL e) ee _.
Program Canonical Structure KES_SKfst ee :=
  @mkKEsplit (tExp6 (Efst ee)) SKfst ee _.
Program Canonical Structure KES_SKsnd ee :=
  @mkKEsplit (tExp7 (Esnd ee)) SKsnd ee _.
Program Canonical Structure KES_SKinl ee :=
  @mkKEsplit (tExp8 (Einl ee)) SKinl ee _.
Program Canonical Structure KES_SKinr ee :=
  @mkKEsplit (tExp9 (Einr ee)) SKinr ee _.
Program Canonical Structure KES_SKcase ee x e1 e2 :=
  @mkKEsplit (tExp10 (Ecase ee x e1 e2)) (SKcase x e1 e2) ee _.
Program Canonical Structure KES_SKroll ee :=
  @mkKEsplit (tExp11 (Eroll ee)) SKroll ee _.
Program Canonical Structure KES_SKunroll ee :=
  @mkKEsplit (tExp12 (Eunroll ee)) SKunroll ee _.
Program Canonical Structure KES_SKappR ee (ve: val) :=
  @mkKEsplit (tExp13 (Eapp ve ee)) (SKappR ve) ee _.
Program Canonical Structure KES_SKappL ee e :=
  @mkKEsplit (tExp14 (Eapp ee e)) (SKappL e) ee _.
Program Canonical Structure KES_SKinst ee :=
  @mkKEsplit (tExp15 (Einst ee)) SKinst ee _.
Program Canonical Structure KES_SKpack ee :=
  @mkKEsplit (tExp16 (Epack ee)) SKpack ee _.
Program Canonical Structure KES_SKunpack ee x e :=
  @mkKEsplit (tExp17 (Eunpack x ee e)) (SKunpack x e) ee _.
Program Canonical Structure KES_SKnew ee :=
  @mkKEsplit (tExp18 (Enew ee)) SKnew ee _.
Program Canonical Structure KES_SKread ee :=
  @mkKEsplit (tExp19 (Eread ee)) SKread ee _.
Program Canonical Structure KES_SKwriteR ee (ve: val) :=
  @mkKEsplit (tExp20 (Ewrite ve ee)) (SKwriteR ve) ee _.
Program Canonical Structure KES_SKwriteL ee e :=
  @mkKEsplit (tExp21 (Ewrite ee e)) (SKwriteL e) ee _.



(** * Properties. *)


Lemma is_value_val: forall {e}
    (X: is_value e),
  e = e_of (mkval _ X).
Proof.
  auto.
Qed.


(** ** Properties about [app_cont]. *)

Lemma app_cont_r_reverse: forall K e,
  app_cont_r K e = app_cont (rev K) e.
Proof. 
 intros; unfold app_cont, app_cont_r.
 rewrite <-fold_left_rev_right, rev_involutive; auto.
Qed.

Lemma app_cont_reverse: forall K e,
  app_cont K e = app_cont_r (rev K) e.
Proof. 
 intros; rewrite <-(rev_involutive K) at 1; rewrite app_cont_r_reverse; auto.
Qed.


(** ** Properties about substitutions. *)

Lemma tmext_vid: forall x,
  [Evar x//x]>>vid = vid.
Proof.
  i; extensionality a; unfold tmext; des_ifs.
Qed.

Lemma subclo_weakly: forall sub,
  subst_closed sub -> subst_weakly_closed sub.
Proof.
  by ii; apply H in FV.
Qed.

Lemma weakclo_tmext: forall sub x ex
    (CLOSED: subst_weakly_closed sub)
    (CLOEX: forall a, fv_exp ex a -> a = x),
  subst_weakly_closed [ex//x]>>sub.
Proof.
  unfold tmext; ii; destruct (@eqdec _); subst; eauto.
Qed.

Lemma subexp_agree: forall (e: exp) sub sub'
    (H: forall x (IN: fv_exp e x), sub x = sub' x),
  subst_exp sub e = subst_exp sub' e.
Proof.
  induction e; s; i
  ; try erewrite IHe; try erewrite IHe1; try erewrite IHe2; try erewrite IHe3
  ; eauto 10; i; unfold tmext; repeat destruct (@eqdec _); eauto 10.
Qed.

Lemma subexp_vid : forall (e: exp),
  subst_exp vid e = e.
Proof. 
  induction e; ss; i; try rewrite !tmext_vid
  ; try setoid_rewrite IHe; try setoid_rewrite IHe1
  ; try setoid_rewrite IHe2; try setoid_rewrite IHe3
  ; eauto 10.
Qed.

Lemma subexp_id_closed: forall (e: exp) sub
    (H: forall x, ~fv_exp e x),
  subst_exp sub e = e.
Proof. 
  i; rewrite <-(subexp_vid e) at 2; apply subexp_agree.
  i; exfalso; eauto.
Qed.

Lemma subexp_closed_weakly: forall e sub x
    (SUB: subst_weakly_closed sub)
    (CLO: forall a, ~ fv_exp (sub x) a),
  ~ fv_exp (subst_exp sub e) x.
Proof.
  unfold subst_weakly_closed; induction e; i; ss; try by des; eauto.
  - assert (FH0:=FH); apply SUB in FH0; subst; eauto.
  - des; [eauto| |].
    + eapply IHe2 in FH0; [eauto|eauto|]
      ; unfold tmext; i; destruct (@eqdec _); subst; eauto.
    + eapply IHe3 in FH0; [eauto|eauto|]
      ; unfold tmext; i; destruct (@eqdec _); subst; eauto.
  - des; eapply IHe in FH1; [eauto|eauto|]
    ; unfold tmext; i; repeat (destruct (@eqdec _); subst; eauto).
  - des; [eauto|].
    des; eapply IHe2 in FH0; [eauto|eauto|]
    ; unfold tmext; i; destruct (@eqdec _); subst; eauto.
Qed.

Lemma subst_closed_tmext: forall e x subst
    (NFV: forall a, ~ fv_exp e a)
    (CLO: subst_closed subst),
  subst_closed [e//x]>>subst.
Proof.
  ii; case (LEM (x = x0)); i; subst.
  - rewrite tmext_eq in FH; eauto.
  - rewrite tmext_neq in FH; eauto.
Qed.

Lemma subexp_closed: forall e sub
    (CLO: subst_closed sub),
  forall x, ~ fv_exp (subst_exp sub e) x.
Proof.
  i; eapply subexp_closed_weakly in FH; eauto.
  by ii; eapply CLO in FV.
Qed.

Local Ltac tmext_case x y := 
  let EQ := fresh "EQ" in let NEQ := fresh "NEQ" in let X := fresh "X" in
  (destruct (eqdec x y) as [EQ|NEQ] eqn: X; subst
   ; [rewrite !tmext_shadow, !tmext_vid, !subexp_vid
     |rewrite !(tmext_swap NEQ), !tmext_vid]; eauto using weakclo_tmext).

Lemma subexp_tmext_r: forall e sub x ex (CLOSED: forall a, ~ fv_exp ex a),
  subst_exp [ex//x]>>sub e = subst_exp sub (subst_exp [ex//x]>>vid e).
Proof.
  induction e; s; i
  ; try by try rewrite IHe; try rewrite IHe1
           ; try rewrite IHe2; try rewrite IHe3; eauto.
  - unfold tmext; destruct (@eqdec _); [rewrite subexp_id_closed|]; eauto.
  - f_equal; eauto; tmext_case x x0.
  - f_equal; tmext_case f x0; tmext_case x x0.
  - f_equal; eauto; tmext_case x x0.
Qed.

Lemma subexp_tmext_l: forall e sub x ex (CLOSED: subst_weakly_closed sub),
   subst_exp [ex//x]>>sub e = 
   subst_exp [ex//x]>>vid (subst_exp [Evar x//x]>>sub e).
Proof.
  induction e; s; i
  ; try by try rewrite IHe; try rewrite IHe1
           ; try rewrite IHe2; try rewrite IHe3; eauto.
  - unfold tmext; destruct (eqdec _ _) eqn: X.
    + subst; ss; rewrite X; eauto.
    + erewrite subexp_agree; [by rewrite subexp_vid|].
      i; apply CLOSED in IN; subst; rewrite X; eauto.
  - f_equal; eauto; tmext_case x x0.
  - f_equal; tmext_case f x0; tmext_case x x0.
  - f_equal; eauto; tmext_case x x0.
Qed.

Lemma subexp_tmext_l2: forall e sub f ef x ex
    (CLOSUB: subst_weakly_closed sub)
    (CLOEX: forall a, ~ fv_exp ex a)
    (CLOEF: forall a, ~ fv_exp ef a),
  subst_exp [ex//x]>>[ef//f]>>sub e =
  subst_exp [ex//x]>>[ef//f]>>vid (subst_exp [Evar x//x]>>[Evar f//f]>>sub e).
Proof.
  i; destruct (eqdec f x) as [EQ|NEQ] eqn: X; subst.
  - rewrite !tmext_shadow, <- subexp_tmext_l; eauto.
  - Local Ltac _auto := by eapply weakclo_tmext; ss; i; exfalso; eauto.
    rewrite subexp_tmext_l; [|by _auto].
    rewrite tmext_swap; [|by eauto].
    rewrite (subexp_tmext_l e); [|by _auto].
    rewrite tmext_swap; [|by eauto].
    rewrite (@subexp_tmext_l _ (tmext f ef vid)); [|by _auto].
    rewrite <-(tmext_swap NEQ _ ef (Evar x)), tmext_vid; eauto.
Qed.

Lemma subst_exp_fv_exp: forall e subst y Y
    (FV: forall x (FV: fv_exp e x) (FVE: fv_exp (subst x) y), Y y : Prop)
    (FVE: fv_exp (subst_exp subst e) y),
  Y y.
Proof.
  induction e; vauto; ss; eauto; i; des
  ; try (by eapply IHe1; eauto)
  ; try (by eapply IHe2; eauto)
  ; try (by eapply IHe3; eauto).
  - eapply (IHe2 _ _ (Y \1/ (fun z => z = x))) in FVE0; des; vauto.
    i; destruct (LEM (x = x0)); subst.
    + rewrite tmext_eq in FVE1; clarify.
    + rewrite tmext_neq in FVE1; eauto 10.
  - eapply (IHe3 _ _ (Y \1/ (fun z => z = x))) in FVE0; des; vauto.
    i; destruct (LEM (x = x0)); subst.
    + rewrite tmext_eq in FVE1; clarify.
    + rewrite tmext_neq in FVE1; eauto 10.
  - eapply (IHe _ _ (Y \1/ (fun z => z = x))) in FVE1; des; vauto.
    i; destruct (LEM (x = x0)); subst.
    + rewrite tmext_eq in FVE2; clarify.
    + rewrite tmext_neq in FVE2; eauto 10.
      destruct (LEM (f = x0)); subst.
      * rewrite tmext_eq in FVE2; clarify.
      * rewrite tmext_neq in FVE2; eauto 10.
  - eapply (IHe2 _ _ (Y \1/ (fun z => z = x))) in FVE0; des; vauto.
    i; destruct (LEM (x = x0)); subst.
    + rewrite tmext_eq in FVE1; clarify.
    + rewrite tmext_neq in FVE1; eauto 10.
Qed.

Lemma subst_exp_closed: forall e subst y
    (FV: forall x (FV: fv_exp e x), ~ fv_exp (subst x) y),
  ~ fv_exp (subst_exp subst e) y.
Proof.
  i; eapply subst_exp_fv_exp; eauto.
Qed.


(** ** Properties about reduction. *)

Lemma red_ctx: forall evt env K ha (ea:exp) hb (eb:exp)
    (RED: reduce_ok evt env ha ea hb eb),
  reduce_ok evt env ha (app_cont K ea) hb (app_cont K eb).
Proof.
  intros; rewrite !app_cont_reverse.
  generalize (rev K); clear K; intro K.
  ginduction K; intros; simpl; eauto.
Qed. 
Hint Resolve red_ctx.

Lemma red_sctx: forall evt env ha ke hb (eb:exp)
    (RED: reduce_ok evt env ha (body_of ke) hb eb),
  reduce_ok evt env ha (term_of ke) hb (app_scont (scont_of ke) eb) .
Proof.
  i; destruct ke as [kea SK ea EQ]; ss; rewrite EQ; eauto.
Qed.
Hint Resolve red_sctx.


(** ** Properties about typing. *)

Lemma mhastype_compose: forall G GA GB ma mb
    (MHTA: mhastype G ma GA)
    (MHTB: mhastype (G ++ GA) mb GB),
  mhastype G (ma ++ mb) (GA ++ GB).
Proof.
  i; destruct MHTA, MHTB; econs; eauto using is_thunkenv_app.
  clear FLI FLO FLI0 FLO0.
  ginduction MHT; i; try rewrite app_nil_r in *; auto.
  rewrite cons_app, <-app_ass in MHT0; ss; eauto.
Qed.

Lemma hastype_tsubst: forall F G ty e a tya
    (HT: hastypepre F G e ty)
    (LC: lc_typ 0 tya),
  hastypepre
    (map (fun x => (fst x, tsubst_name [tya//a]>>tid (snd x))) F) 
    (map (fun x => (fst x, tsubst_name [tya//a]>>tid (snd x))) G) 
    e (tsubst_name [tya//a]>>tid ty).
Proof.
  i; ginduction e; ss; i; depdes HT
  ; try by try eapply IHe in HT; eauto; try eapply IHe1 in HT1; eauto
         ; try eapply IHe2 in HT2; eauto; try eapply IHe3 in HT3; eauto
         ; ss; econs; eauto using lc_tsubname, lc_tmext_ty.
  - econs; induction G; ss.
    destruct a0 as (y,tyy); ss.
    destruct eq_nat_dec; [depdes H|]; eauto.
  - econs; induction F; ss.
    destruct a0 as (y,tyy); ss.
    destruct eq_nat_dec; [depdes H|]; eauto.
  - eapply IHe in HT; eauto.
    rewrite tsub_name_bvar in HT; s; eauto using lc_tmext_ty.
  - eapply IHe in HT; eauto.
    rewrite tsub_name_bvar; s; eauto using lc_tmext_ty.
  - eapply (IHe _ _ _ aa (Tname (frnam_l (Tname a::tya::ty::(map snd F ++ map snd G))))) in HT.
    2: by eauto.
    rewrite tsub_name_bvar in HT; eauto using lc_tmext_ty.
    ss; rewrite tsubname_id in HT; eauto.
    do 2 (rewrite tsubnameL_id in HT; eauto). 
    rewrite tmext_eq in HT.
    apply (IHe _ _ _ a tya) in HT; eauto.
    rewrite tsub_name_bvar in HT; eauto using lc_tmext_ty.
    rewrite tsubname_id in HT; [|s; i; omega].
    assert (X:= frnam_l_app (map snd F) (map snd G)).
    econs; [by eauto| |].
    + rewrite map_map; setoid_rewrite <- (map_map (@snd termvar typ)).
      i; eapply (frnam_l_tsub (ty::map snd F)) in H1.
      eapply frnam_l_big in H1; ss; omega.
    + rewrite map_map; setoid_rewrite <- (map_map (@snd termvar typ)).
      i; eapply (frnam_l_tsub (ty::map snd G)) in H1.
      eapply frnam_l_big in H1; ss; omega.
  - eapply IHe in HT; eauto.
    rewrite tsub_name_bvar; s; eauto using lc_tsubname, lc_tmext_ty.
  - eapply IHe in HT; eauto.
    rewrite tsub_name_bvar in HT; ss; eauto using lc_tsubname, lc_tmext_ty.
  - eapply (IHe2 _ _ _ aa
            (Tname (frnam_l (Tname a::tya::tya0::tyb::(map snd F ++ map snd G))))) in HT2.
    2: by eauto.
    ss; rewrite tsub_name_bvar in HT2; eauto using lc_tmext_ty.
    ss; rewrite tsubname_id in HT2; eauto.
    do 2 (rewrite tsubnameL_id in HT2; eauto).
    rewrite tmext_eq in HT2.
    rewrite tsubname_id in HT2; eauto.
    apply (IHe2 _ _ _ a tya) in HT2; eauto.
    ss; rewrite tsub_name_bvar in HT2; eauto using lc_tmext_ty.
    rewrite tsubname_id in HT2; [|s; i; omega].
    eapply IHe1 in HT1; eauto; ss.
    assert (X:= frnam_l_app (map snd F) (map snd G)).
    econs; eauto using lc_tsubname, lc_tmext_ty.
    + rewrite map_map; setoid_rewrite <- (map_map (@snd termvar typ)).
      i; eapply (frnam_l_tsub (tya0::tyb::map snd F)) in H2.
      eapply frnam_l_big in H2; ss; omega.
    + rewrite map_map; setoid_rewrite <- (map_map (@snd termvar typ)).
      i; eapply (frnam_l_tsub (tya0::tyb::map snd G)) in H2.
      eapply frnam_l_big in H2; ss; omega.
Qed.

Lemma hastype_weakening: forall F F' G G' e ty
    (HT: hastypepre F G e ty)
    (LEF: forall x tyx, list_assoc eqdec x F = Some tyx ->
                      list_assoc eqdec x F' = Some tyx)
    (LEG: forall x tyx, list_assoc eqdec x G = Some tyx ->
                      list_assoc eqdec x G' = Some tyx),
  hastypepre F' G' e ty.
Proof.
  i; ginduction e; ss; i; depdes HT
  ; try by try eapply IHe in HT; eauto; try eapply IHe1 in HT1; eauto
         ; try eapply IHe2 in HT2; eauto; try eapply IHe3 in HT3; eauto
         ; ss; econs; eauto using lc_tsubname, lc_tmext_name.
  - econs; eauto; [eapply IHe2|eapply IHe3]; eauto
    ; s; i; destruct eq_nat_dec; eauto.
  - econs; eapply IHe; eauto; s; i; repeat destruct eq_nat_dec; eauto.
  - eapply (hastype_tsubst aa) in HT.
    2: by instantiate (1 := Tname (frnam_l (ty :: (map snd F' ++ map snd G')))); eauto.
    ss; do 2 (rewrite tsubnameL_id in HT; eauto).
    rewrite tsub_name_bvar in HT; eauto using lc_tmext_name.
    ss; rewrite tmext_eq in HT.
    rewrite tsubname_id in HT; eauto.
    eapply IHe in HT; eauto.
    assert (X:= frnam_l_app (map snd F') (map snd G')).
    econs; [by eauto| |]; i; eapply frnam_l_big in FH; ss; omega.
  - eapply (hastype_tsubst aa) in HT2.
    2: by instantiate (1 := Tname (frnam_l (tya :: tyb :: (map snd F' ++ map snd G')))); eauto.
    ss; do 2 (rewrite tsubnameL_id in HT2; eauto).
    rewrite tsub_name_bvar in HT2; eauto using lc_tmext_name.
    ss; rewrite tmext_eq in HT2.
    rewrite tsubname_id in HT2; eauto.
    rewrite tsubname_id in HT2; eauto.
    assert (X:= frnam_l_app (map snd F') (map snd G')).
    eapply (IHe2 _ _ _ ((x,_) :: G')) in HT2; [|by eauto |].
    + eapply HTunpack with (aa := frnam_l (tya :: tyb :: (map snd F' ++ map snd G')))
      ; eauto; i; apply frnam_l_big in FH; ss; omega.
    + s; i; destruct eq_nat_dec; subst; eauto.
Qed.

Lemma hastype_swap: forall F G G' x y tyx tyy ty e2
    (HT: hastypepre F (G ++ [(x, tyx); (y, tyy)] ++ G') e2 ty)
    (NE: ~ x = y),
  hastypepre F (G ++ [(y, tyy); (x, tyx)] ++ G') e2 ty.
Proof.
  i; eapply hastype_weakening; eauto.
  clear - NE; induction G.
  - s; i; destruct eq_nat_dec; subst. 
    + depdes H; destruct eq_nat_dec; subst; intuition.
    + destruct eq_nat_dec; eauto.
  - s; i; destruct a, eq_nat_dec; eauto.
Qed.

Lemma hastype_fv: forall F G e ty x
    (HT: hastypepre F G e ty)
    (FV: fv_exp e x),
  In x (map fst G).
Proof.
  i; ginduction HT; i; ss; subst; try by des; eauto.
  - induction G; i; [done|].
    ss; destruct a, eq_nat_dec; subst; eauto.
  - by des; eauto; [apply IHHT2 in FV0|apply IHHT3 in FV0]; des; subst; eauto.
  - by des; eauto; apply IHHT in FV1; des; subst; eauto.
  - by des; eauto; apply IHHT2 in FV0; des; subst; eauto.
Qed.

Lemma hastype_subst: forall F x tyx G ex e ty
    (HTE: hastypepre F ((x, tyx) :: G) e ty)
    (HTX: hastypepre F [] ex tyx),
  hastypepre F G (subst_exp [ex//x]>>vid e) ty.
Proof.
  i; ginduction e; i; depdes HTE; ss; eauto.
  - destruct eq_nat_dec.
    + subst; depdes H; rewrite tmext_eq; eauto. 
      by eapply hastype_weakening; eauto.
    + rewrite tmext_neq; eauto.
  - destruct (eq_nat_dec x0 x).
    + subst; rewrite !tmext_shadow, !tmext_vid, !subexp_vid.
      econs; eauto; eapply hastype_weakening; eauto
      ; s; i; destruct eq_nat_dec; subst; eauto.
    + rewrite tmext_swap; eauto; rewrite tmext_vid.
      econs; eauto; [eapply IHe2|eapply IHe3]; eauto
      ; eapply (hastype_swap []); eauto.
  - destruct (eq_nat_dec x0 f).
    + subst; rewrite !tmext_shadow, !tmext_vid, !subexp_vid.
      econs; eauto; eapply hastype_weakening; eauto.
      repeat (s; i; destruct eq_nat_dec; subst; eauto).
    + rewrite (@tmext_swap _ _ _ f x0); eauto; rewrite tmext_vid.
      destruct (eq_nat_dec x0 x).
      * subst; rewrite !tmext_shadow, !tmext_vid, !subexp_vid.
        econs; eauto; eapply hastype_weakening; eauto.
        repeat (s; i; destruct eq_nat_dec; subst; eauto).
      * rewrite tmext_swap; eauto; rewrite tmext_vid. 
        econs; eauto; eapply IHe; eauto.
        eapply (hastype_swap []); eauto.
        s; eapply (hastype_swap [_]); eauto.
  - econs; eauto; ss; intuition.
  - destruct (eq_nat_dec x0 x).
    + subst; rewrite !tmext_shadow, !tmext_vid, !subexp_vid.
      econs; eauto. 
      * instantiate (1:= aa); s; intuition.
      * s; intuition.
      * eapply hastype_weakening; eauto
        ; s; i; destruct eq_nat_dec; subst; eauto.
    + rewrite tmext_swap; eauto; rewrite tmext_vid.
      econs; eauto.
      * instantiate (1:= aa); s; intuition.
      * s; intuition.
      * eapply IHe2; eauto.
        eapply (hastype_swap []); eauto.
Qed.


 
