(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** A library of total and partial maps, represented as functions. *)

Require Import Arith Bool List FunctionalExtensionality ClassicalChoice.
Require Import basis myclassical vgtac countable.
Require Import ssr.



Set Implicit Arguments.
Generalizable Variables X.
Global Set Bullet Behavior "Strict Subproofs".



(** * Total maps.

    Will be used mainly to represent substitutions.
*)


(** ** Main definitions. *)

(** The type of total maps from [X] to [Y]. *)
Definition tmap `{M:EqDec X} (Y:Type) := X -> Y.

(** Extending or updating a total map with a single mapping. *)
Definition tmext `{M:EqDec X} {Y:Type} (loc: X) (v: Y) (h: tmap Y) : tmap Y :=
  fun l => if eqdec l loc then v else h l.
Notation "[ y // x ] >> tm" := (tmext x y tm) (at level 0, right associativity).


(** ** Properties of [tmext]. *)

Lemma tmext_eq `{M:EqDec X}: forall A sub x (a: A),
  ([a//x] >> sub) x = a.
Proof.
  by unfold tmext; i; rewrite eqdec_true.
Qed.

Lemma tmext_neq `{M:EqDec X}: forall A sub x (a: A) y (NEQ: y <> x),
  ([a//x] >> sub) y = sub y.
Proof.
  by unfold tmext; i; rewrite (eqdec_false NEQ).
Qed.

Lemma tmext_shadow `{M:EqDec X}: forall A sub x (a a': A),
  [a'//x] >> [a//x] >> sub = [a'//x] >> sub.
Proof.
  i; extensionality y; unfold tmext; destruct eqdec; eauto.
Qed.







Lemma tmext_lift `{M: EqDec X}: forall A B (f: A -> B) sub x (a: A),
  [f a//x]>>(fun x' => f (sub x')) =
  fun x' => f ([a//x]>>sub x').
Proof.
  i; extensionalities; unfold tmext; destruct eqdec; subst; auto.
Qed.

Lemma tmext_swap `{M:EqDec X}: forall A x x' (NEQ: x <> x') sub (ex ex': A),
  [ex//x] >> [ex'//x'] >> sub = [ex'//x'] >> [ex//x] >> sub.
Proof.
  i; extensionality y; unfold tmext; destruct eqdec; eauto.
  by subst; destruct eqdec; subst; eauto.
Qed.

Lemma tmext_self `{M: EqDec X}: forall x A (sub: tmap A),
  sub = [sub x//x]>>sub.
Proof.
  i; extensionalities; unfold tmext; des_if; subst; auto.
Qed.


(** ** Tactics. *)

Ltac tmext_simpl :=
  repeat match goal with |- context[[?a//?x]>>?sub ?y] =>
    rewrite (@tmext_neq _ _ _ sub x a y); [|by vauto]
  end;
  repeat match goal with |- context[[?a//?x]>>?sub ?x] =>
    rewrite (@tmext_eq _ _ _ sub x a)
  end.

Ltac tmext_simplH H :=
  repeat match goal with HH: context[[?a//?x]>>?sub ?y] |- _ =>
    check_equal HH H; rewrite (@tmext_neq _ _ _ sub x a y) in H; [|by vauto]
  end;
  repeat match goal with HH: context[[?a//?x]>>?sub ?x] |- _ =>
    check_equal HH H; rewrite (@tmext_eq _ _ _ sub x a) in H
  end.

(** Remove shadowed binding.

    This rewrites a map of the form
      [y//x_i]>>[y_n//x_n]>>...>>[y_1//x_1]>>tm
    (where x_1 to x_n are syntactically distinct) to
      [y//x_i]>>[y_n//x_n]>>...>>[y_i+1//x_i+1]>>
      [y_i-1//x_i-1]>>...>>[y_1//x_1]>>tm
*)
Ltac tmext_uniq :=
  match goal with |- context[ [_//?r]>>_ ] =>
    repeat (rewrite -(@tmext_swap _ _ _ r); [|by vauto]);
    rewrite (tmext_shadow _ r)
  end.



(** * Partial maps.

    Will be used mainly to represent heaps and stacks.
*)


(** ** Main definitions. *)

(** The type of partial maps from X to Y. *)
Definition pmap `{M:EqDec X} (Y:Type) := option (X -> option Y).


Definition pmadd `{M:EqDec X} {Y:Type} (ha hb: X -> option Y) : X -> option Y :=
  fun l => match ha l with
           | Some v => Some v
           | None => hb l
           end.



Definition pmdisj `{M:EqDec X} {Y:Type} (oha ohb: pmap Y) : Prop :=
  match oha, ohb with
  | Some ha, Some hb => forall l, ha l = None \/ hb l = None
  | _, _ => False
  end.


(** Validity (well-definedness). *)
Definition pmvalid `{M:EqDec X} {Y:Type} (h: pmap Y) : Prop :=
  ~ h = None.

(** Disjoint union of two partial maps.  Note that this is a function. *)
Definition pmplus `{M:EqDec X} {Y:Type} (oha ohb: pmap Y) : pmap Y :=
  match oha, ohb with
  | Some ha, Some hb =>
    if LEM (pmdisj oha ohb) then Some (pmadd ha hb) else None
  | _, _ => None
  end.
Arguments pmplus _ _ _ _  _: simpl never.
Infix "##" := pmplus (at level 60, right associativity).

(** Empty partial map. *)
Definition pmemp `{M:EqDec X} {Y:Type} : pmap Y := Some (fun l => None).

(** Singleton partial map. *)
Definition pmcell `{M:EqDec X} {Y:Type} (loc: X) (v: Y) : pmap Y :=
  Some (fun l => if eqdec l loc then Some v else None).
Notation "{ x |-> y }" := (pmcell x y).

(** Reading from a partial map. *)
Definition pmread `{M:EqDec X} {Y:Type} (h: pmap Y) l (v: Y) : Prop :=
  <<VLD: pmvalid h>> /\ <<RD: exists h', h = pmcell l v ## h'>>.

(** Finiteness. *)
Definition pmfinite `{M:EqDec X} {Y:Type} (h: pmap Y) : Prop :=
  exists findom, forall l v (RD: pmread h l v), In l findom.

(** Consecutive chain of cells (finite) *)
Fixpoint pmcells Y (start: nat) (vs: list Y) := 
  match vs with
  | nil => pmemp
  | v :: vs' => pmcell start v ## pmcells (1+start) vs'
  end.
Arguments pmcells [Y] _ _ : simpl never.
Notation "{ x |=> ys }" := (pmcells x ys).


(** ** Properties. *)
Section pmap_properties.

  Variable X: Type.
  Variable M: EqDec X.
  Variable Y: Type.


  (** *** Extensionality. *)

  Lemma pmvalid_some: forall (hp: pmap Y)
      (VLD: pmvalid hp),
    exists h, hp = Some h.
  Proof.
    i; unfold pmvalid in VLD; edestruct hp; vauto.
  Qed.

  Lemma pmread_some: forall (h: X -> option Y) x y,
    pmread (Some h) x y <-> h x = Some y.
  Proof.
    split; i.
    - rrdes H; unfold pmplus in RD; destruct h'; vauto; des_ifH RD; vauto.
      ss; clarify; specialize (p x); unfold pmadd; des_ifH p; des; vauto.
    - split; [ii; vauto|].
      exists (Some (fun z => if eqdec z x then None else h z)).
      unfold pmplus; s; des_if.
      + f_equal; extensionality z; s; unfold pmadd; des_if; vauto.
      + exfalso; apply f; i; des_if; vauto.
  Qed.

  Lemma pm_extensionality: forall (ha hb: pmap Y)
      (VLDa: pmvalid ha)
      (VLDb: pmvalid hb)
      (RD: forall x y, pmread ha x y <-> pmread hb x y),
    ha = hb.
  Proof.
    i; apply pmvalid_some in VLDa; apply pmvalid_some in VLDb; des; subst.
    f_equal; extensionality l.
    destruct (h0 l) eqn:E.
    - eapply pmread_some in E; apply RD, pmread_some in E; auto.
    - destruct (h l) eqn:E'; auto.
      eapply pmread_some in E'; apply RD, pmread_some in E'; vauto.
  Qed.


  (** *** Properties of [pmplus]. *)

  Lemma pmplus_assoc: forall (ha hb hc: pmap Y),
    ha ## hb ## hc = (ha ## hb) ## hc.
  Proof.
    i; destruct ha, hb, hc; auto; [|unfold pmplus in *; des_if; auto].
    unfold pmplus, pmdisj, pmadd; des_ifs; f_equal; extensionalities
    ; [by destruct (o H); eauto| | | |] 
    ; exfalso; apply f; i.
    - specialize (o4 l); des; eauto.
      specialize (o3 l); rewrite o4 in o3; destruct (o l); eauto. 
    - specialize (o3 l); des; eauto.
      destruct (o0 l); eauto.
    - specialize (o5 l); des; eauto.
      specialize (o4 l); rewrite o5; destruct (o l); eauto.
    - specialize (o4 l); des; eauto.
      specialize (o3 l); rewrite o4 in o3; eauto.
  Qed.

  Lemma pmplus_comm: forall (ha hb: pmap Y),
    ha ## hb = hb ## ha.
  Proof.
    i; destruct ha, hb; auto.
    unfold pmplus; des_ifs
    ; [|M|M]; Mdo by exfalso; apply f; ii; specialize (p l); des; eauto.
    f_equal; extensionality l; unfold pmadd.
    specialize (p0 l); des; destruct (o l), (o0 l); eauto; vauto.
  Qed.

  Lemma pmplus_swap: forall (ha hb hc: pmap Y),
    ha ## hb ## hc = hb ## ha ## hc.
  Proof.
    i; rewrite ->pmplus_assoc, (pmplus_comm ha), <-pmplus_assoc; auto.
  Qed.

  Lemma pmplus_emp: forall (h: pmap Y),
    pmemp ## h = h.
  Proof. 
    i; destruct h; eauto.
    unfold pmplus; destruct (LEM _); ss.
    exfalso; apply f; eauto.
  Qed.

  Lemma pmemp_plus: forall (ha hb: pmap Y)
      (E: pmemp = ha ## hb),
    ha = pmemp /\ hb = pmemp.
  Proof.
    i; destruct ha as [|ha], hb as [|hb]; ss; unfold pmplus in E
    ; des_ifH E; vauto; inversion E.
    split; unfold pmemp; f_equal; extensionality x; assert (A:=(eq_fun H0 x))
    ; ss; specialize (p x); unfold pmadd in A; destruct (o x); vauto.
  Qed.

  Lemma pmvalid_plusD: forall (hx hy: pmap Y)
      (PV: pmvalid (hx ## hy)),
    pmvalid hx /\ pmvalid hy.
  Proof.
    i; destruct hx, hy; done.
  Qed.

  Lemma pmplus_inj: forall (h ha hb: pmap Y)
      (E: h ## ha = h ## hb)
      (VLD: pmvalid (h ## ha)),
    ha = hb.
  Proof.
    i; hexploit pmvalid_plusD; eauto; []; i; des.
    destruct h, ha, hb; vauto; [|r in X1; vauto].
    f_equal; extensionality l.
    dup VLD; rewrite E in VLD0.
    unfold pmplus in E, VLD, VLD0; unfold pmvalid in VLD, VLD0.
    des_ifH E; des_ifH E; vauto.
    hexploit (eq_fun H0 l); i.
    unfold pmadd in X2.
    specialize (p l); specialize (p0 l).
    destruct (o l) in *; eauto.
    des; vauto; rewrite p; eauto.
  Qed.

  (** Split a partial map in two according to a given predicate [P]. *)
  Lemma pmplus_from_pred: forall (h: pmap Y) (P: X -> Prop),
    exists hP hNP,
      h = hP ## hNP /\
      <<RDP: forall x y, pmread hP x y <-> pmread h x y /\ P x>> /\
      <<RDNP: forall x y, pmread hNP x y <-> pmread h x y /\ ~ P x>>.
  Proof.
    i; destruct h; last first.
    { exists None, None; split; auto.
      split; ii; split; i; des; rrdes H VLD; vauto.
    }
    set (hPf := fun l => if LEM (P l) then o l else None).
    set (hNPf := fun l => if LEM (P l) then None else o l).
    exists (Some hPf), (Some hNPf); unfold hPf, hNPf; split.
    - unfold pmplus; des_if.
      + f_equal; extensionalities; unfold pmadd.
        des_if; auto; destruct (o H); auto.
      + exfalso; apply f; ii; des_if; auto.
    - split; i; split; i; des.
      + apply pmread_some in H; des_ifH H; vauto.
        split; [apply pmread_some|]; auto.
      + apply pmread_some; des_if; vauto.
        eapply pmread_some; auto.
      + apply pmread_some in H; des_ifH H; vauto.
        split; [apply pmread_some|]; auto.
      + apply pmread_some; des_if; vauto.
        eapply pmread_some; auto.
  Qed.


  (** *** Properties of [pmvalid]. *)

  Lemma pmvalid_emp:
    pmvalid (@pmemp _ _ Y).
  Proof.
    done.
  Qed.

  Lemma pmvalid_cell: forall (loc: X) (v: Y),
    pmvalid {loc |-> v}.
  Proof.
    done.
  Qed.

  Lemma pmvalid_plus_l: forall (hx hy: pmap Y) 
      (PV: pmvalid (hx ## hy)),
    pmvalid hx.
  Proof.
    i; destruct hx, hy; done.
  Qed.

  Lemma pmvalid_plus_r: forall (hx hy: pmap Y) 
      (PV: pmvalid (hx ## hy)),
    pmvalid hy.
  Proof. 
    i; rewrite pmplus_comm in PV; eapply pmvalid_plus_l; eauto.
  Qed.

  Lemma pmvalid_plus_cell: forall l v v' (h: pmap Y)
      (PV: pmvalid ({l |-> v} ## h)),
    pmvalid ({l |-> v'} ## h).
  Proof.
    ii; apply PV; clear PV; destruct h; auto.
    unfold pmplus in *; ss; des_ifs.
    exfalso; apply f; i; specialize (o0 l0).
    by des_ifs; des; eauto.
  Qed.                          

  Lemma pmvalid_plus_plus: forall (ha hb hc: pmap Y)
      (PVab: pmvalid (ha ## hb))
      (PVac: pmvalid (ha ## hc))
      (PVbc: pmvalid (hb ## hc)),
    pmvalid (ha ## hb ## hc).
  Proof.
    i; apply pmvalid_some in PVab; apply pmvalid_some in PVac.
    apply pmvalid_some in PVbc; des; unfold pmvalid in *; i.
    unfold pmplus in FH; destruct ha, hb, hc; vauto; des_ifH FH.
    - des_ifH FH; vauto; apply f; ii.
      unfold pmplus in PVab; unfold pmplus in PVac.
      des_ifH PVab; des_ifH PVac; vauto.
      specialize (p0 l); specialize (p1 l); des; vauto.
      right; unfold pmadd; rewrite p0; eauto.
    - apply f; unfold pmplus in PVbc; des_ifH PVbc; vauto.
  Qed.

  Lemma pmvalid_cell_cell: forall (la lb: X) (va vb: Y),
    pmvalid ({la |-> va} ## {lb |-> vb}) <->
    la <> lb.
  Proof.
    split; ii; unfold pmplus in *; ss.
    - subst; des_ifs; vauto; eauto.
      specialize (o lb); des_ifs; des; vauto.
    - unfold pmplus in *; ss; des_ifs.
      apply f; i; destruct eqdec, eqdec; subst; s; eauto.
      subst; exfalso; eauto.
  Qed.

  
  
  
  
  
  
  
  

  Corollary pmvalid_plus_plusD: forall (ha hb hc: pmap Y)
      (VLD: pmvalid (ha ## hb ## hc)),
    pmvalid (ha ## hb) /\
    pmvalid (ha ## hc) /\
    pmvalid (hb ## hc).
  Proof.
    i; splits.
    - rewrite pmplus_assoc in VLD; eauto using pmvalid_plus_l.
    - rewrite ->(pmplus_comm hb), pmplus_assoc in VLD; eauto using pmvalid_plus_l.
    - eauto using pmvalid_plus_r.
  Qed.

  Lemma pmvalid_plus_read: forall (ha hb: pmap Y)
      (VLDa: pmvalid ha)
      (VLDb: pmvalid hb)
      (RD: forall x ya yb
               (RDa: pmread ha x ya)
               (RDa: pmread hb x yb),
             False),
    pmvalid (ha ## hb).
  Proof.
    ii; destruct ha, hb; vauto; unfold pmplus in FH.
    des_ifH FH; vauto; eapply f.
    ii; destruct (o l) eqn:E; destruct (o0 l) eqn:E0; vauto.
    exfalso; apply pmread_some in E; apply pmread_some in E0; eauto.
  Qed.

  Lemma pmvalid_cell_plus: forall h (x: X) (y: Y)
      (VLD: pmvalid h)
      (NEW: forall x' y' (RD: pmread h x' y'), x' <> x),
    pmvalid ({x |-> y} ## h).
  Proof.
    i; destruct h; vauto.
    unfold pmplus; s; des_if; unfold pmvalid; vauto.
    exfalso; apply f; i; des_if; auto; subst; right.
    destruct (o x) eqn:E; auto; specialize (NEW x y0).
    exfalso; apply NEW; auto; apply pmread_some; auto.
  Qed.


  (** *** Properties of [pmcell]. *)

  Lemma pmcell_inj: forall (x x': X) (y y': Y)
      (E: {x |-> y} = {x' |-> y'}),
    x = x' /\ y = y'.
  Proof.
    i; inversion E; assert (A:=eq_fun H0 x); ss.
    
    do 2 (des_ifH A; vauto).
  Qed.

  Lemma pmcell_uniq: forall (x: X) (y y': Y) hp hp'
      (VLD: pmvalid ({x |-> y} ## hp))
      (E: {x |-> y} ## hp = {x |-> y'} ## hp'),
    y = y' /\ hp = hp'.
  Proof.
    i; unfold pmplus, pmvalid in *; ss.
    destruct hp, hp'; vauto; des_ifH E; try des_ifH E; vauto; split.
    - assert (A:=eq_fun H0 x); unfold pmadd in A; des_ifH A; vauto.
    - f_equal; extensionalities; specialize (o1 H); specialize (o2 H).
      assert (A:=eq_fun H0 H); unfold pmadd in A; des_ifH A; vauto.
      des; vauto; rewrite ->o1, o2; auto.
  Qed.
  
  Lemma pmcell_distinct: forall (x x': X) (y y': Y) hp hp'
      (VLD: pmvalid ({x |-> y} ## hp))
      (E: {x |-> y} ## hp = {x' |-> y'} ## hp')
      (NE: x <> x'),
    exists hp'', hp = {x' |-> y'} ## hp'' /\ hp' = {x |-> y} ## hp''.
  Proof.
    i; rrdes VLD; destruct hp, hp'; vauto; assert (VLD':=VLD); rewrite E in VLD'.
    unfold pmplus in E, VLD, VLD'; ss; des_ifH E; des_ifH E; vauto.
    exists (Some (fun z => if eqdec z x then None else if eqdec z x' then None else o z)).
    unfold pmcell, pmplus; split.
    - des_if; [|by exfalso; apply f; ii; des_if; vauto; des_if; vauto].
      f_equal; extensionalities.
      specialize (o1 H); specialize (o2 H).
      assert (A:=eq_fun H0 H); ss; unfold pmadd in *.
      des_if; vauto; des_ifH A; des; vauto.
    - des_if; [|by exfalso; apply f; ii; des_if; vauto; des_if; vauto].
      f_equal; extensionalities.
      specialize (o1 H); specialize (o2 H).
      assert (A:=eq_fun H0 H); ss; unfold pmadd in *.
      des_if; vauto; des_ifH A; des; vauto.
  Qed.

  Lemma pmcell_plus: forall (x x': X) (y y': Y) hp
      (E: {x |-> y} = {x' |-> y'} ## hp),
    hp = pmemp.
  Proof.
    i; destruct hp; vauto; unfold pmemp; f_equal; extensionality z.
    apply NNPP; ii.
    unfold pmcell, pmplus, pmadd in E; des_ifH E; clarify.
    assert (A:=eq_fun H1 z); ss; des_ifH A.
    - des_ifH A; vauto.
      + assert (B:=p x'); des_ifH B; des; vauto.
      + assert (B:=eq_fun H1 x'); ss; do 2 des_ifH B; vauto.
    - des_ifH A; try apply H; vauto.
  Qed.


  (** *** Properties of [pmread]. *)

  Lemma pmread_valid: forall (h: pmap Y) x y
      (RD: pmread h x y),
    pmvalid h.
  Proof.
    i; rrdes RD; auto.
  Qed.

  Lemma pmread_cell: forall (x: X) (y: Y),
    pmread {x |-> y} x y.
  Proof.
    split; eauto using pmvalid_cell.
    exists pmemp; rewrite ->pmplus_comm, pmplus_emp; auto.
  Qed.

  Lemma pmread_cellD: forall (x x': X) (y y': Y)
      (RD: pmread {x |-> y} x' y'),
    x = x' /\ y = y'.
  Proof.
    i; rrdes RD; destruct h'; vauto.
    assert (E: Some o = pmemp); [eapply pmcell_plus; eauto|].
    rewrite ->E, pmplus_comm, pmplus_emp in RD0; eauto using pmcell_inj.
  Qed.

  Lemma pmread_uniq: forall (h: pmap Y) x ya yb
      (RDa: pmread h x ya)
      (RDb: pmread h x yb),
    ya = yb.
  Proof.
    i; rrdes RDa RDb; subst; eapply pmcell_uniq; eauto.
  Qed.

  Lemma pmread_plusD: forall (ha hb: pmap Y) x y
      (RD: pmread (ha ## hb) x y),
    pmread ha x y \/ pmread hb x y.
  Proof.
    i; rrdes RD.
    assert (VLDa:=VLD); apply pmvalid_plus_l in VLDa.
    assert (VLDb:=VLD); apply pmvalid_plus_r in VLDb.
    destruct (LEM (pmread ha x y)); vauto.
    destruct (LEM (pmread hb x y)); vauto.
    exfalso; assert (VLD':=VLD); apply pmvalid_some in VLD'; des.
    des; unfold pmplus in VLD'; destruct ha, hb; vauto; des_ifH VLD'; vauto.
    destruct (LEM (o x = Some y)); des; [|destruct (LEM (o0 x = Some y))]; des.
    - apply f, pmread_some; auto.
    - apply f0, pmread_some; auto.
    - rewrite RD0 in VLD; assert (A:=pmvalid_plus_r _ VLD).
      eapply pmvalid_some in A; des; subst.
      unfold pmplus in RD0; do 2 (des_ifH RD0; vauto); s in RD0.
      inversion RD0; assert (A:=(eq_fun H0 x)).
      unfold pmadd in A; des_ifH A; vauto; destruct (o x); vauto.
  Qed.

  Lemma pmread_plus: forall (ha hb h: pmap Y) x y
      (RD: pmread ha x y)
      (E: ha ## hb = h)
      (VLD: pmvalid h),
    pmread h x y.
  Proof.
    i; rrdes RD; subst.
    econs; [|rewrite <-pmplus_assoc]; eauto.
  Qed.

  Lemma pmread_plus_cell: forall a b u v w (h: pmap Y)
     (RD: pmread ({a |-> u} ## h) b w)
     (NE: a <> b),
    pmread ({a |-> v} ## h) b w.
   Proof.
     i; dup RD; apply pmread_valid in RD.
     apply pmread_plusD in RD0; des.
     - eapply pmread_cellD in RD0; des; subst; vauto.
     - eapply pmread_plus; eauto using pmvalid_plus_cell.
       erewrite pmplus_comm; auto.
   Qed.

  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  

  Lemma pmread_empD: forall x (y: Y),
    ~ pmread pmemp x y.
  Proof.
    ii; rrdes FH; apply pmemp_plus in RD; des; inversion RD.
    assert (A:=(eq_fun H0 x)); ss; destruct (eqdec x x); vauto.
  Qed.

  Lemma pmemp_from_read: forall (h: pmap Y)
      (VLD: pmvalid h)
      (RD: forall x y, ~ pmread h x y),
    h = pmemp.
  Proof.
    i; assert (VLD':=VLD); rrdes VLD'; case_eq h; vauto.
    i; unfold pmemp; f_equal; extensionality x.
    case_eq (o x); vauto; i; exfalso; eapply (RD x y).
    econs; eauto; exists (Some (fun z => if eqdec z x then None else o z)).
    unfold pmplus; s; des_if.
    - f_equal; extensionality z; s; unfold pmadd; des_if; vauto.
    - exfalso; apply f; i; des_if; vauto.
  Qed.

  Lemma pmplus_from_read: forall (h h': pmap Y)
      (RD: forall x v (RD: pmread h x v), pmread h' x v)
      (VLD: pmvalid h),
    exists h'', h' = h ## h''.
  Proof.
    i; rrdes VLD; destruct h, h'; vauto; last first.
    { exists None; eauto. }
    exists (Some (fun x => match o0 x with
                             | None => None
                             | Some v =>
                               match o x with
                                 | Some _ => None
                                 | None => Some v
                               end
                           end)).
    unfold pmplus; des_if.
    - f_equal; extensionality x; unfold pmadd.
      destruct (o x) eqn:E.
      + eapply pmread_some, RD, pmread_some; auto.
      + destruct (o0 x); eauto.
    - exfalso; apply f; ii.
      destruct (o l); vauto; right; destruct (o0 l); auto.
  Qed.

  Lemma pmvalidD_plus_read: forall (ha hb: pmap Y) x ya yb
      (VLD: pmvalid (ha ## hb))
      (RDa: pmread ha x ya)
      (RDa: pmread hb x yb),
    False.
  Proof.
    i; apply pmvalid_some in VLD; des.
    unfold pmplus in VLD; destruct ha, hb; vauto; des_ifH VLD; vauto.
    apply pmread_some in RDa; apply pmread_some in RDa0.
    specialize (p x); des; vauto.
  Qed.

  (** *** Properties of [pmfinite]. *)

  Lemma pmfinite_emp:
    pmfinite (@pmemp _ _ Y).
  Proof.
    exists nil; i; apply pmread_empD in RD; auto.
  Qed.

  Lemma pmfinite_cell: forall (loc: X) (v: Y),
    pmfinite {loc |-> v}.
  Proof. 
    i; exists [loc]; i; apply pmread_cellD in RD; des; s; auto.
  Qed.

  Lemma pmfinite_plus: forall (ha hb: pmap Y)
      (FINa: pmfinite ha)
      (FINb: pmfinite hb),
    pmfinite (ha ## hb).
  Proof.
    i; rdes FINa; rdes FINb; destruct ha, hb; vauto.
    des_ifs; exists (findom ++ findom0).
    i; specialize (FINa l v); specialize (FINb l v).
    apply pmread_plusD in RD; des; apply in_or_app; [left|right]; auto.
  Qed.

  Lemma pmfinite_plusD: forall (ha hb: pmap Y)
      (VLD: pmvalid (ha ## hb))
      (FIN: pmfinite (ha ## hb)),
    pmfinite ha /\ pmfinite hb.
  Proof.
    i; rrdes FIN; destruct ha, hb; vauto; unfold pmvalid in *; ss.
    split; auto; exists findom; i; eapply FIN; eapply pmread_plus
    ; eauto using pmplus_comm.
  Qed.

  Corollary pmfinite_plus_cell: forall l v v' (h: pmap Y)
      (FIN: pmfinite ({l |-> v} ## h))
      (VLD: pmvalid ({l |-> v} ## h)),
    pmfinite ({l |-> v'} ## h).
  Proof.
    i; apply pmfinite_plusD in FIN; des
    ; auto using pmfinite_plus, pmfinite_cell.
  Qed.

  Lemma pmfinite_alloc: forall (h: pmap Y) y
      (VLD: pmvalid h)
      (FIN: pmfinite h)
      (INF: countably_infinite X),
    exists x,
      pmvalid ({x |-> y} ## h).
  Proof.
    i; rrdes FIN INF.
    set (new := g (S (list_max (map f findom)))).
    assert (~ In new findom).
    {
      i; apply (in_map f), list_max_in in FH.
      cut (f new <> S (list_max (map f findom))); [|omega]; eauto.
    }
    exists new; apply pmvalid_plus_read; eauto using pmvalid_cell.
    i; apply pmread_cellD in RDa; des; apply FIN in RDa0; subst; subst; auto.
  Qed.                          

End pmap_properties.

Hint Resolve pmread_valid.
Hint Resolve pmread_cell.
Hint Resolve pmvalid_emp.
Hint Resolve pmfinite_emp.
Hint Resolve pmfinite_cell.
Hint Resolve pmfinite_plus.

Notation "{ x |-> y }" := (pmcell x y).
Notation "{ x |=> ys }" := (pmcells x ys).


(** *** Some tactics. *)

(** Simplify combinations of [pmemp] and [pmplus].

    Concretely, apply [pmplus_emp] wherever possible (symmetrically).
*)
Ltac nopmemp :=
  repeat match goal with
    | H: context[pmemp ## ?X] |- _ => rewrite (pmplus_emp X) in H
    | |- context[pmemp ## ?X]      => rewrite (pmplus_emp X)
    | H: context[?X ## pmemp] |- _ => rewrite (pmplus_comm X pmemp) in H
    | |- context[?X ## pmemp]      => rewrite (pmplus_comm X pmemp)
  end.

(** Split combinations of [pmvalid] and [pmplus].

    Concretely, apply [pmvalid_plus_c] wherever possible (symmetrically).
*)
Ltac split_pmvalid VLD :=
  guard; unguardH VLD;
  repeat match goal with
    | H: pmvalid (pmplus ?a ?b) |- _
    => apply pmvalid_plusD in H; des
  end; unguard.

(** Split combinations of [pmfinite] and [pmplus].

    Concretely, apply [pmfinite_plusD] wherever possible (symmetrically).
    This also splits [pmvalid] because [pmfinite_plusD] requires validity.
 *)
Ltac split_pmfinite FIN VLD :=
  guard; unguardH FIN; unguardH VLD;
  repeat match goal with
    | H: pmfinite (pmplus ?a ?b) |- _ =>
      apply pmfinite_plusD in H; [|by auto]; des
    | H: pmvalid (pmplus ?a ?b) |- _ =>
      apply pmvalid_plusD in H; des
  end; unguard.

(** Decompose combinations of [pmvalid] and [pmplus].

    Concretely, apply [pmvalid_plus_plusD] wherever possible (symmetrically).
 *)
Ltac decomp_pmvalid VLD :=
  guard; unguardH VLD;
  repeat match goal with
    | H: pmvalid (pmplus ?a (pmplus ?b ?c)) |- _ =>
      apply pmvalid_plus_plusD in H; des
    | H: pmvalid (pmplus (pmplus ?a ?b) ?c) |- _ =>
      rewrite pmplus_comm in H
  end; unguard.

(** Order consecutive cells (in goal). *)
Ltac pmcells_order :=
  repeat match goal with |- context[{S ?n |-> ?x} ## ({?n |-> ?y} ## ?h)] =>
    rewrite (pmplus_swap {S n |-> x} {n |-> y} h)
  end;
  repeat match goal with |- context[{S ?n |-> ?x} ## {?n |-> ?y}] =>
    rewrite (pmplus_comm {S n |-> x} {n |-> y})
  end.

(** Fold chains of consecutive [pmcell]s into single [pmcells] (in goal). *)
Ltac pmcells_fold :=
  repeat match goal with
    | |- context[{?x |-> ?y} ## {S ?x |=> ?ys}]
    => replace ({x |-> y} ## {S x |=> ys}) with {x |=> y :: ys}
         ; [|by unfold pmcells]

    | |- context[{?x |-> ?y} ## {S ?x |=> ?ys} ## ?h]
    => rewrite (pmplus_assoc {x |-> y} {S x |=> ys} h)

    | |- context[{?x |-> ?y} ## {S ?x |-> ?z}]
    => replace ({x |-> y} ## {S x |-> z}) with {x |=> [y; z]}
       ; [|by unfold pmcells; nopmemp]

    | |- context[{?x |-> ?y} ## {S ?x |-> ?z} ## ?h]
    => match h with
         | {S (S x) |-> _} ## _ => fail 1
         | _ => rewrite (pmplus_assoc {x |-> y} {S x |-> z})
       end
  end.                          

(** Fold chains of consecutive [pmcell]s into single [pmcells] (in hyp). *)
Ltac pmcells_foldH H :=
  move H at bottom; repeat match goal with

    | H': context[{?x |-> ?y} ## {S ?x |=> ?ys}] |- _
    => check_equal H' H
       ; replace ({x |-> y} ## {S x |=> ys}) with {x |=> y :: ys} in H'
       ; [|by unfold pmcells]

    | H': context[{?x |-> ?y} ## {S ?x |=> ?ys} ## ?h] |- _
    => check_equal H' H
       ; rewrite (pmplus_assoc {x |-> y} {S x |=> ys} h) in H'

    | H': context[{?x |-> ?y} ## {S ?x |-> ?z}] |- _
    => replace ({x |-> y} ## {S x |-> z}) with {x |=> [y; z]} in H'
       ; [|by unfold pmcells; nopmemp]

    | H': context[{?x |-> ?y} ## {S ?x |-> ?z} ## ?h] |- _
    => match h with
         | {S (S x) |-> _} ## _ => fail 1
         | _ => rewrite (pmplus_assoc {x |-> y} {S x |-> z}) in H
       end
  end.


(** *** Properties of [pmcells]. *)

Lemma pmcells_nil: forall Y n,
  {n |=> @nil Y} = pmemp.
Proof.
  auto.
Qed.

Lemma pmcells_cons: forall Y n (y: Y) ys,
  {n |=> y :: ys} = {n |-> y} ## {S n |=> ys}.
Proof.
  auto.
Qed.

Lemma pmcells_app: forall Y (ys zs: list Y) n,
  {n |=> ys ++ zs} = {n |=> ys} ## {n + length ys |=> zs}.
Proof.
  induction ys; i; s; [by rewrite ->pmplus_emp, plus_0_r|].
  unfold pmcells in *; rewrite <-pmplus_assoc, IHys; do 3 f_equal; omega.
Qed.

Lemma pmcells_sng: forall Y (y: Y) n,
  {n |-> y} = {n |=> [y]}.
Proof.
  i; unfold pmcells; nopmemp; auto.
Qed.

Lemma pmcells_plus_cell: forall {Y} m n (y: Y) ys
    (GE: m > n),
  pmvalid ({n |-> y} ## {m |=> ys}).
Proof.
  i; ginduction ys; i; vauto.
  - rewrite ->pmplus_comm, pmplus_emp; apply pmvalid_cell.
  - unfold pmcells; fold (@pmcells Y); apply pmvalid_plus_plus.
    + apply pmvalid_cell_cell; i; subst; omega.
    + eapply (IHys (S m) n); eauto.
    + eapply (IHys (S m) m); eauto.
Qed.

Lemma pmcells_distinct: forall Y (ys: list Y) y m n h h'
    (VLD: pmvalid ({m |=> ys} ## h))
    (E: {m |=> ys} ## h = {n |-> y} ## h')
    (LE: n < m \/ m + length ys <= n),
  exists h'', h = {n |-> y} ## h'' /\ h' = {m |=> ys} ## h''.
Proof.
  induction ys; i; [by esplits; ss; rewrite ->pmplus_emp in *; eauto|].
  ss; unfold pmcells in *; rewrite <-pmplus_assoc in *.
  apply pmcell_distinct in E; [|by eauto|by omega].
  destruct E as [? []]; subst; apply pmvalid_plusD in VLD.
  destruct VLD; apply IHys in H; [|by eauto|by omega].
  destruct H as [? []]; subst; esplits; try rewrite <-pmplus_assoc; eauto.
Qed.

Lemma pmcells_uniq: forall Y (ys ys': list Y) m h h'
    (VLD: pmvalid ({m |=> ys} ## h))
    (E: {m |=> ys} ## h = {m |=> ys'} ## h')
    (LEN: length ys = length ys'),
  ys = ys' /\ h = h'.
Proof.
  induction ys; destruct ys'; vauto; i.
  - rewrite pmcells_nil in E; nopmemp; eauto.
  - rewrite !pmcells_cons -!pmplus_assoc in E VLD.
    apply pmcell_uniq in E; eauto; des; subst; eauto.
    eapply pmvalid_plusD in VLD; des.
    hexploit IHys; eauto; i; des; subst; eauto.
Qed.

Lemma pmcells_valid: forall {Y} n (ys: list Y),
  pmvalid {n |=> ys}.
Proof.
  i; destruct ys; vauto; unfold pmcells; fold (@pmcells Y). 
  apply pmcells_plus_cell; omega.
Qed.
Hint Resolve pmcells_valid.

Lemma pmcells_finite: forall {Y} n (ys: list Y),
  pmfinite {n |=> ys}.
Proof.
  i; ginduction ys; i; s; eauto using pmfinite_emp; apply pmfinite_plus
  ; auto using pmcells_valid, pmfinite_cell.
Qed.
Hint Resolve pmcells_finite.

Lemma pmcells_read: forall {Y} n m nm (y: Y) ys
    (Ev: nth_error ys n = Some y)
    (E: nm = n + m),
  pmread {m |=> ys} nm y.
Proof.
  induction n; i; subst; (destruct ys; [by ss; omega|]; rewrite pmcells_cons).
  - eapply pmread_plus; [by eapply pmread_cell|by ss; inv Ev|].
    pmcells_fold; apply pmcells_valid.
  - s; replace (S (n + m)) with (n + S m); [|by omega].
    eapply pmread_plus; [by apply IHn; ss; eauto; omega|by apply pmplus_comm|].
    pmcells_fold; apply pmcells_valid.
Qed.

Lemma pmcells_readD: forall x Y ys x' (y: Y)
    (RD: pmread {x |=> ys} x' y),
  <<LE: x <= x' < x + length ys>> /\ <<NTH: nth_error ys (x' - x) = Some y>>.
Proof.
  i; ginduction ys; i.
  - rewrite pmcells_nil in RD; exfalso; eauto using pmread_empD.
  - rewrite pmcells_cons in RD.
    apply pmread_plusD in RD; des.
    + apply pmread_cellD in RD; des; subst.
      repeat split; s; eauto; [omega|by rewrite <-minus_n_n; s; eauto].
    + apply IHys in RD; des; repeat split; s; eauto; [omega|omega|].
      destruct (x' - x) eqn:E; s; eauto; [omega|].
      rewrite <-NTH; replace n with (x' - S x); eauto; omega.
Qed.

Lemma pmcells_plus_cells: forall {Y} (la lb: list Y) na nb
    (LE: na + length la <= nb),
  pmvalid ({na |=> la} ## {nb |=> lb}).
Proof.
  induction la; s; i; [by rewrite pmplus_emp; eauto using pmcells_valid|].
  unfold pmcells in *; rewrite <-pmplus_assoc.
  apply pmvalid_plus_plus; try (apply pmcells_plus_cell; omega).
  eapply IHla; eauto; omega.
Qed.

Lemma pmvalid_plus_cells: forall Y vs vs' (h: pmap Y) n
    (PV: pmvalid ({n |=> vs} ## h))
    (LEN: length vs = length vs'),
  pmvalid ({n |=> vs'} ## h).
Proof.
  induction vs; i; destruct vs'; vauto.
  rewrite ->pmcells_cons, <-pmplus_assoc in *.
  unfold pmcells; fold (@pmcells Y); apply pmvalid_plus_plus.
    + rewrite <-pmcells_cons; apply pmcells_valid.
    + rewrite pmplus_swap in PV; apply pmvalid_plusD in PV; des.
      eauto using pmvalid_plus_cell.
    + apply pmvalid_plusD in PV; des; eauto.
Qed.

Lemma pmvalid_cells_plus: forall Y n (ys: list Y) findom hp
    (VLD: pmvalid hp)
    (FIN: forall l y (RD: pmread hp l y), In l findom)
    (LE: forall l (IN: In l findom), l < n),
  pmvalid ({n |=> ys} ## hp).
Proof.
  i; ginduction ys; i; [by unfold pmcells; nopmemp; eauto|].
  rewrite ->pmcells_cons, <-pmplus_assoc; apply pmvalid_cell_plus.
  - eapply IHys; eauto; i; eapply lt_trans; eauto.
  - i; apply pmread_plusD in RD; des.
    + cut (x' > n); [by omega|]; clear -RD; ginduction ys; i.
      * unfold pmcells in RD; vauto; apply pmread_empD in RD; vauto.
      * rewrite pmcells_cons in RD; apply pmread_plusD in RD; des.
          by apply pmread_cellD in RD; des; omega.
          eapply gt_trans; [by eapply IHys; eauto; omega|by omega].
    + cut (x' < n); eauto; omega.
Qed.

Lemma pmfinite_plus_cells: forall Y vs vs' (h: pmap Y) n n'
    (VLD: pmvalid ({n |=> vs} ## h))
    (FIN: pmfinite ({n |=> vs} ## h)),
  pmfinite ({n' |=> vs'} ## h).
Proof.
  i; apply pmfinite_plusD in FIN; eauto; des.
  eapply pmfinite_plus; eauto.
Qed.

























