(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Omega Program ClassicalChoice.
Require Import ssr vgtac basis myclassical lang_common lang_spec.
Require Export lang_tgt_syntax.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".



Definition enc: instr -> word :=
  enum.

Lemma enc_inj: forall i j, enc i = enc j -> i = j.
Proof.
  exact enum_inj.
Qed.


Definition regfile := @tmap reg _ word.
Definition stack := @pmap word _ word.
Definition heap := @pmap word _ word.
Structure machine :=
  mkmach {mpc: word; mrf: regfile; mst: stack; mhp: heap}.


(** Operational Semantics *)


Definition add_offset n soff := 
  match soff with
    | o_pos m => m + n
    | o_neg m => n - m
  end.

Inductive eval rf st h : operand -> word -> Prop :=
| ev_imm w : eval rf st h (o_imm w) w
| ev_reg r : eval rf st h (o_reg r) (rf r)
| ev_sind r soff w  
    (RD: pmread st (add_offset (rf r) soff) w) :
  eval rf st h (o_sind r soff) w
| ev_hind r soff w  
    (RD: pmread h (add_offset (rf r) soff) w) :
  eval rf st h (o_hind r soff) w
.
Hint Constructors eval.

Inductive store pc rf : stack -> heap -> operand -> word -> machine -> Prop :=
| sto_reg st h r w :
  store pc rf st h (o_reg r) w (mkmach (1+pc) [w//r]>>rf st h)
| sto_sind st h r soff v w :
  store pc rf ({add_offset (rf r) soff |-> v} ## st) h (o_sind r soff) w
    (mkmach (1+pc) rf ({add_offset (rf r) soff |-> w} ## st) h)
| sto_hind st h r soff v w :
  store pc rf st ({add_offset (rf r) soff |-> v} ## h) (o_hind r soff) w
    (mkmach (1+pc) rf st ({add_offset (rf r) soff |-> w} ## h))
.
Hint Constructors store.


(** We assume a deterministic (but otherwise arbitrary) memory allocator. *)

Definition valid_allocator (alloc: heap -> nat -> nat) :=
  forall h n (FIN: pmfinite h) (VLD: pmvalid h),
    pmvalid ({alloc h n |=> repeat 0 n} ## h).

Definition raw_heap := list (word * word).

Axiom raw_alloc: raw_heap -> word -> word.
Axiom raw_alloc_allocator: forall rh i n
    (LT: i < n),
  lookup (i + raw_alloc rh n) rh = None.

Notation cid := constructive_indefinite_description.

(** Lift [raw_alloc] to the notion of heap used by the operational semantics. *)
Definition alloc (h: heap) (n: word) : word :=
  match LEM (pmfinite h), LEM (pmvalid h) with
    | left FIN, left VLD =>
      let findom := proj1_sig (cid _ FIN) in
      let hf := proj1_sig (cid _ (pmvalid_some VLD)) in
      let f l := match hf l with Some v => [(l, v)] | _ => [] end in
      raw_alloc (flatten (map f findom)) n
    | _, _ => 42 
  end.

Lemma alloc_valid:
  valid_allocator alloc.
Proof.
  unfold valid_allocator, alloc; i; des_ifs.
  destruct (cid _ p); destruct (cid _ (pmvalid_some p0)).
  subst.

  set (rh := flatten _).

  assert (forall l v, pmread (Some x0) l v <-> lookup l rh = Some v).
  {
    split; i.
    - dup H; apply i in H; apply in_split_first in H; des.
      apply pmread_some in H0.
      unfold rh; subst; s; clear -H0 H1; induction l1.
      { s; rewrite H0 -cons_app; des_ifs. }
      exploit IHl1; [by ss; eauto|]; i; s.
      rewrite -x1; s; des_ifs.
      rewrite -cons_app list_assoc_cons_neq; auto.
      ii; subst; ss; eauto.
    - apply list_assoc_someD in H; apply flatten_inD in H; des.
      apply in_map_iff in INb; des.
      des_ifs; s in INa; des; vauto.
      apply pmread_some; auto.
  }

  apply pmvalid_plus_read; auto; []; i; apply H in RDa0.
  apply pmcells_readD in RDa; des.
  assert (X := @raw_alloc_allocator rh (x1 - raw_alloc rh n) n).
  hexploit X; [rewrite repeat_length in LE0; omega|].
  replace (x1 - _ + _) with x1; [|omega]; i; vauto.
Qed.



Inductive step_ok : event -> machine -> machine -> Prop := 
| Sjmp: forall pc rf st h addr pc'
    (INS: pmread h pc (enc (i_jmp addr)))
    (EVAL: eval rf st h addr pc'),
  step_ok etau (mkmach pc rf st h) (mkmach pc' rf st h)
| Sjz: forall pc rf st h r addr
    (INS: pmread h pc (enc (i_jnz r addr)))
    (EVALV: rf r = 0),
  step_ok etau (mkmach pc rf st h) (mkmach (1+pc) rf st h)
| Sjnz: forall pc rf st h r n addr pc'
    (INS: pmread h pc (enc (i_jnz r addr)))
    (EVALV: rf r = S n)
    (EVALA: eval rf st h addr pc'),
  step_ok etau (mkmach pc rf st h) (mkmach pc' rf st h)
| Sld: forall pc rf st h r src w
    (INS: pmread h pc (enc (i_ld r src)))
    (EVAL: eval rf st h src w),
  step_ok etau (mkmach pc rf st h) (mkmach (1+pc) [w//r]>>rf st h) 
| Ssto: forall pc rf st h tgt src m
    (INS: pmread h pc (enc (i_sto tgt src)))
    (UPD: store pc rf st h tgt (rf src) m),
  step_ok etau (mkmach pc rf st h) m
| Slpc: forall pc rf st h r
    (INS: pmread h pc (enc (i_lpc r))),
  step_ok etau (mkmach pc rf st h) (mkmach (1+pc) [pc//r]>>rf st h) 
| Sopr: forall pc rf st h op tgt arga argb va vb m
    (INS: pmread h pc (enc (i_opr op tgt arga argb)))
    (EVALa: eval rf st h arga va)
    (EVALb: eval rf st h argb vb)
    (UPD: store pc rf st h (o_reg tgt) (bosem op va vb) m),
  step_ok etau (mkmach pc rf st h) m
| Salc: forall pc rf st h r rsize loc
    (INS: pmread h pc (enc (i_alc r rsize)))
    (ALC: loc = alloc h (rf rsize)),
  step_ok etau (mkmach pc rf st h)
    (mkmach (1+pc) [loc//r]>>rf st ({loc |=> repeat 0 (rf rsize)} ## h))
| Sinp: forall pc rf st h r n
    (INS: pmread h pc (enc (i_inp r))),
  step_ok (ein n) (mkmach pc rf st h) (mkmach (1+pc) [n//r]>>rf st h) 
| Sout: forall pc rf st h r
    (INS: pmread h pc (enc (i_out r))),
  step_ok (eout (rf r)) (mkmach pc rf st h) (mkmach (1+pc) rf st h)
.

Definition mkerr m := mkmach m.(mpc) m.(mrf) m.(mst) None.

Inductive tgt_step : event -> machine -> machine -> Prop :=
| Sok:  forall evt m m'
          (NZ: ~ m.(mpc) = 0)
          (ST: step_ok evt m m'),
        tgt_step evt m m'
| Serr: forall m
          (NZ: m.(mpc) <> 0)
          (ST: forall evt m', ~ step_ok evt m m'),
        tgt_step etau m (mkerr m)
.
Hint Constructors tgt_step.

Definition tgt_halted m := m.(mpc) = 0. 


Definition o_zero := o_pos 0.
Definition o_one  := o_pos 1.
Definition o_mone := o_neg 1.



(** Loading and Linking *)


Definition segment := (word * list word) %type.
Definition module := word -> list word -> list (label * segment).

Definition msize (codes: list (list word)) :=
  let sizes := map (fun code => S (length code)) codes in
  fold_right plus 0 sizes.


Definition link (m1 m2: module) : module :=
  fun laddr imports =>
    let lsegs1 := m1 laddr imports in
    let size := msize (map (snd <*> snd) lsegs1) in
    let lsegs2 := m2 (laddr + size) (imports ++ map (fst <*> snd) lsegs1) in
    lsegs1 ++ lsegs2.

Structure config :=
  mkconf {cpc: eoption word; crf: eoption regfile; cst: stack; chp: heap}.

Definition realizes (c: config) (m: machine) : Prop :=
  <<PC: c.(cpc) = Just m.(mpc)>> /\
  <<RF: c.(crf) = Just m.(mrf)>> /\
  <<ST: c.(cst) = m.(mst)>> /\ <<SV: pmvalid m.(mst)>> /\
  <<HP: c.(chp) = m.(mhp)>> /\ <<HV: pmvalid m.(mhp)>> /\ <<FIN: pmfinite m.(mhp)>>.

Definition loadseg (seg: segment) : heap * heap :=
  ({fst seg |-> S (fst seg)}, {S (fst seg) |=> snd seg}).

Fixpoint loadgh (segs: list segment) : heap :=
  match segs with
    | nil => pmemp
    | seg :: segs' => fst (loadseg seg) ## loadgh segs'
  end.

Fixpoint loadlh (segs: list segment) : heap :=
  match segs with
    | nil => pmemp
    | seg :: segs' => snd (loadseg seg) ## loadlh segs'
  end.

Definition tgt_load (m: module) : option machine :=
  let lsegs := m 0 nil in
  match lookup MAIN lsegs with
    | Some (laddr, code) =>
      Some {| mpc := S laddr; mrf := [laddr//r_clo]>>(fun _ => 0);
              mst := Some (fun _ => Some 0);
              mhp := loadgh (map snd lsegs) ## loadlh (map snd lsegs) |}
    | _ =>
      None
  end.                          

Definition tgt_wfmod (m: module) exports :=
  <<DJ: forall laddr imports la lb (NEQ: la <> lb)
               sega (INa: In (la, sega) (m laddr imports))
               segb (INb: In (lb, segb) (m laddr imports)),
          pmvalid (fst (loadseg sega) ## snd (loadseg sega) ##
                   fst (loadseg segb) ## snd (loadseg segb))>> /\
  <<BND: forall laddr imports l w ws (IN: In (l, (w, ws)) (m laddr imports)),
           laddr <= w /\
           w + length (ws) < laddr + msize (map (snd <*> snd) (m laddr imports))>> /\
  <<UNQ: forall laddr imports l
                na sega (INa: nth_error (m laddr imports) na = Some (l, sega))
                nb segb (INb: nth_error (m laddr imports) nb = Some (l, segb)),
           na = nb>> /\
  <<OUT: forall laddr imports, exports = map fst (m laddr imports)>>.

  



Lemma msize_app: forall xs ys,
  msize (xs ++ ys) = msize xs + msize ys.
Proof.
  induction xs; i; auto.
  rewrite <-app_comm_cons; s; unfold msize; s. 
  rewrite <-plus_assoc; do 2 f_equal; apply IHxs.
Qed.

Lemma msize_link: forall ma mb laddr imports,
  msize (map (snd <*> snd) (link ma mb laddr imports)) =
  msize (map (snd <*> snd) (ma laddr imports)) +
  msize (map (snd <*> snd) (mb (laddr + msize (map (snd <*> snd) (ma laddr imports)))
                               (imports ++ map (fst <*> snd) (ma laddr imports)))).
Proof.
  i; unfold link; rewrite map_app; apply msize_app.
Qed.


Theorem tgt_link_wfmod: forall ma mb outa outb
    (WFa: tgt_wfmod ma outa) (WFb: tgt_wfmod mb outb)
    (DJ: forall lbl (INa: In lbl outa) (INb: In lbl outb), False),
  tgt_wfmod (link ma mb) (outa ++ outb).
Proof.
  i; rrdes WFa WFb; rr; splits; i.
  - apply in_app_or in INa; apply in_app_or in INb; des.
    + eapply DJ0; eauto.
    + destruct sega, segb; apply BND in INb; apply BND0 in INa; des; s.
      rewrite pmplus_assoc pmplus_comm; pmcells_fold.
      apply pmcells_plus_cells; s; omega.
    + destruct sega, segb; apply BND in INa; apply BND0 in INb; des; s.
      rewrite pmplus_assoc; pmcells_fold.
      apply pmcells_plus_cells; s; omega.
    + eapply DJ1; eauto.
  - apply in_app_or in IN; des; [apply BND in IN|apply BND0 in IN]; des; split
    ; auto; try rewrite msize_link; omega.
  - apply nth_error_app in INa; apply nth_error_app in INb; des.
    + eapply UNQ; eauto.
    + exfalso.
      eapply nth_error_someD, (in_map fst) in INa0; rewrite <-OUT0 in INa0.
      eapply nth_error_someD, (in_map fst) in INb; rewrite <-OUT in INb.
      eapply DJ; eauto.
    + exfalso.
      eapply nth_error_someD, (in_map fst) in INa; rewrite <-OUT in INa.
      eapply nth_error_someD, (in_map fst) in INb0; rewrite <-OUT0 in INb0.
      eapply DJ; eauto.
    + exploit UNQ0; [exact INa0|exact INb0|]; i; instantiate; omega.
  - unfold link; erewrite map_app, <-OUT, OUT0; eauto.
Qed.




Section LangSpecImpl.

  Definition tgt_val := word.
  Definition tgt_cont := word.
  Definition tgt_conf := config.
  Definition tgt_mach := machine.
  Definition tgt_module := module.
  Definition tgt_anchor := word.

  Instance Enumerable_tgt_val: Enumerable tgt_val :=
    {| enum := enum; enum_inj := enum_inj |}.

  

  Definition tgt_vload (m: module) laddr (imports: list (label * word)) lbl v :=
    
    exists code, lookup lbl (m laddr (map snd imports)) = Some (v, code).

  Definition tgt_cload (m: module) laddr (imports: list (label * word)) (cg cl: config) :=
    
    exists hp_imports rf,
      <<GPC: cg.(cpc) = Nothing>> /\
      <<GRF: cg.(crf) = Just rf>> /\
      <<GST: rf r_sp = 0 /\ forall a, pmread cg.(cst) a 0>> /\
      <<GHP: cg.(chp) = hp_imports ## loadgh (map snd (m laddr (map snd imports)))>> /\
      <<GDJ: pmvalid cg.(chp)>> /\
      <<GDOM: forall loc, In loc (map snd imports) <-> exists v, pmread hp_imports loc v>> /\
      <<LPC: cl.(cpc) = Nothing>> /\
      <<LRF: cl.(crf) = Nothing>> /\
      <<LST: cl.(cst) = pmemp>> /\
      <<LHP: cl.(chp) = loadlh (map snd (m laddr (map snd imports)))>> /\
      <<LDJ: pmvalid (hp_imports ## cl.(chp))>>.

  Definition tgt_cplus (c c': tgt_conf) : tgt_conf :=
    match c, c' with {| cpc := cpc; crf := crf; cst := cst; chp := chp |},
                     {| cpc := cpc'; crf := crf'; cst := cst'; chp := chp' |} =>
      {| cpc := eoplus cpc cpc'
       ; crf := eoplus crf crf'
       ; cst := cst ## cst'
       ; chp := chp ## chp'
      |}
    end.

  Definition tgt_cemp :=
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := pmemp |}.

  Definition tgt_core c := exists pc,
    c.(cpc) = Just pc /\ c.(crf) = Nothing /\ c.(cst) = pmemp /\ c.(chp) = pmemp.
  Definition tgt_extra c := c.(cpc) = Nothing /\ c.(crf) = Nothing.
  Definition tgt_real := realizes.

  Lemma store_st: forall pc rf st hp tgt src m
      (UPD: store pc rf st hp tgt src m)
      (VLD: pmvalid (mst m)),
    pmvalid st.
  Proof.
    i; inv UPD; eauto using pmvalid_plus_cell.
  Qed.

  Lemma store_hp: forall pc rf st hp tgt src m
      (UPD: store pc rf st hp tgt src m)
      (FIN: pmfinite (mhp m))
      (VLD: pmvalid (mhp m)),
    pmfinite hp /\ pmvalid hp.
  Proof.
    i; inv UPD; split; eauto using pmvalid_plus_cell.
    ss; split_pmfinite FIN VLD; eauto.
  Qed.

  Definition tgt_lsi: LangSpec.
    refine
      {| val := tgt_val
       ; cont := tgt_cont
       ; conf := tgt_conf
       ; mach := tgt_mach
       ; lang_spec.module := tgt_module
       ; anchor := tgt_anchor
       ; vload := tgt_vload
       ; cload := tgt_cload
       ; cplus := tgt_cplus
       ; cemp := tgt_cemp
       ; halted := tgt_halted
       ; core := tgt_core
       ; extra := tgt_extra
       ; lang_spec.step := tgt_step
       ; real := tgt_real
      |}.
    - split.
      + i; destruct m; unfold tgt_halted in *; ss.
        inversion FH; vauto.
      + i; r; s; apply NNPP; intro NV.
        destruct (classic (exists evt m', step_ok evt m m')); des; eauto.
    - destruct a, b; unfold tgt_cplus; s.
      f_equal; eauto using pmplus_comm, eoplus_comm.
    - i; destruct a, b, c; s; f_equal; eauto using pmplus_assoc, eoplus_assoc.
    - i; destruct a; s; destruct cpc0, crf0; rewrite !pmplus_emp; auto.
    - i; apply RL; i; rrdes FH0; destruct c; ss; clarify.
      inv ST; [|by rrdes HV; vauto].
      inv ST0; ss; try (by eapply (FH (mkconf _ _ _ _)); econs; esplits; ss; eauto).
      + hexploit store_st; eauto; i; hexploit store_hp; eauto; i; des.
        eapply (FH (mkconf _ _ _ _)); econs; esplits; ss; eauto.
      + hexploit store_st; eauto; i; hexploit store_hp; eauto; i; des.
        eapply (FH (mkconf _ _ _ _)); econs; esplits; ss; eauto.
      + split_pmfinite FIN HV.
        eapply (FH (mkconf _ _ _ _)); rr; esplits; ss; eauto.
    - i; apply NNPP; ii; inversion ST; vauto; inversion ST0; vauto
      ; apply NE; i; rrdes FH; ss; clarify
      ; eapply (E (mkconf _ _ _ _)); rr; esplits; ss; eauto.
    - i; invc COa COb RLa RLb; des; subst.
      destruct m, coa, ca, cob, cb; ss; noemp; clarify.
      destruct cpc1, cpc3; noemp; subst; vauto.
    - ss; clarify.
    - i; destruct c, c'; rrdes XTR XTR'; ss; subst; rr; eauto.
    - set (dummy := {| mpc := 0; mrf := fun _ => 0; mst := pmemp; mhp := pmemp |}).
      i; destruct c.
      destruct cpc0; [|M|M]
      ; Mdo (by exists dummy; i; destruct cf; destruct cpc0; inv XTR; rrdes RL; vauto).
      destruct crf0; [|M|M]
      ; Mdo (by exists dummy; i; destruct cf; destruct crf0; inv XTR; rrdes RL; vauto).
      econs; instantiate (1 := {| mpc := _; mrf := _; mst := _; mhp := _ |}). 
      destruct cf; ss; clarify; eauto.
      i; inv XTR; rrdes RL; rewrite <-HP, <-ST in *; ss.
      split_pmfinite FIN HV; split_pmvalid SV; rr; esplits; ss; eauto.
      
      
      
      
      
    - i; destruct m; unfold tgt_halted in *; ss; des.
      destruct cf; rrdes XTR; ss; clarify.
      destruct mm; rrdes RL; unfold tgt_cplus in *; destruct c'; ss; clarify.
      destruct cpc0, crf0; rrdes RLm; ss; clarify.
  Defined.

  Definition tgt_link := link.

End LangSpecImpl.

Hint Unfold tgt_core.






Lemma loadgh_app: forall a b, loadgh (a ++ b) = loadgh a ## loadgh b.
Proof.
  induction a; i; s; noemp; auto.
  destruct a; rewrite <-pmplus_assoc; s; f_equal; eauto.
Qed.

Lemma loadlh_app: forall a b, loadlh (a ++ b) = loadlh a ## loadlh b.
Proof.
  induction a; i; s; noemp; auto.
  destruct a; s; rewrite <-pmplus_assoc; f_equal; eauto.
Qed.


Lemma loadgh_loadlh_valid_aux:
  forall m exports (WF: tgt_wfmod m exports) laddr imports
         junk (VLD: pmvalid junk)
         (DJ: forall l seg (IN: In (l, seg) (m laddr imports)),
                pmvalid (junk ## fst (loadseg seg) ## snd (loadseg seg))),
    pmvalid (junk ## loadgh (map snd (m laddr imports)) ##
                     loadlh (map snd (m laddr imports))).
Proof.
  i; rrdes WF; specialize (DJ0 laddr imports); specialize (UNQ laddr imports).
  generalize dependent junk; induction (m laddr imports); s; i
  ; noemp; vauto.
  destruct a; simpl map; unfold loadgh, loadlh; fold loadgh loadlh; s; i.
  rewrite <-pmplus_assoc, (pmplus_swap (loadgh _)).
  rewrite 2!pmplus_assoc -(pmplus_assoc junk).
  apply IHl; i.
  - eapply DJ0; try right; eauto.
  - apply eq_add_S; eapply UNQ; eauto.
  - eapply DJ; eauto.
  - rewrite pmplus_comm; apply pmvalid_plus_plus; [M|M|]
    ; (Mdo rewrite pmplus_comm); try (eapply DJ; eauto).
    rewrite pmplus_comm -pmplus_assoc.
    eapply (DJ0 l1 l0); [|right|left]; eauto; ii; subst.
    apply nth_error_some in IN; des; cut (0 = S n); vauto; eapply UNQ; s; eauto.
Qed.

Lemma loadgh_loadlh_valid:
  forall m exports (WF: tgt_wfmod m exports) laddr imports,
    pmvalid (loadgh (map snd (m laddr imports)) ##
             loadlh (map snd (m laddr imports))).
Proof.
  i; assert (X := loadgh_loadlh_valid_aux WF laddr imports (@pmvalid_emp _ _ _)).
  noemp; apply X; i; noemp; unfold loadseg; s.
  pmcells_fold; apply pmcells_valid.
Qed.


Lemma loadgh_finite: forall segs,
    pmfinite (loadgh segs).
Proof.
  induction segs; s; auto.
Qed.

Lemma loadlh_finite: forall segs,
    pmfinite (loadlh segs).
Proof.
  induction segs; s; auto.
Qed.

Lemma loadgh_closure_aux: forall (f: label) lsegs v code
    (IN: lookup f lsegs = Some (v, code)),
  exists h, loadgh (map snd lsegs) = {v |-> S v} ## h.
Proof.
  induction lsegs; i; vauto.
  destruct a as [? []]; apply list_assoc_cons_or in IN; des; clarify.
  - s; eauto.
  - hexploit IHlsegs; eauto; []; i; des.
    s; econs; rewrite X pmplus_swap; eauto.
Qed.

Lemma loadgh_closure: forall f m laddr imports oenv v code
    (IN: lookup f (m laddr imports) = Some (v, code))
    (WF: tgt_wfmod m oenv),
  pmread (loadgh (map snd (m laddr imports))) v (S v).
Proof.
  i; hexploit (loadgh_loadlh_valid WF laddr imports); i; split_pmvalid X.
  apply loadgh_closure_aux in IN; des; eapply pmread_plus; eauto.
Qed.

Lemma tgt_cload_finite: forall m laddr imports cg cl
    (LD: tgt_cload m laddr imports cg cl),
  pmfinite cg.(chp) /\ pmfinite cl.(chp).
Proof.
  i; rrdes LD; rewrite GHP LHP.
  split; [apply pmfinite_plus|]; eauto using loadgh_finite, loadlh_finite.
  exists (map snd imports); i; apply GDOM; eauto.
Qed.




Lemma eval_uniq: forall rf st hp op wa wb
    (EVa: eval rf st hp op wa)
    (EVb: eval rf st hp op wb),
  wa = wb.
Proof.
  i; inv EVa; inv EVb; eauto using (@pmread_uniq nat).
Qed.

Lemma store_uniq: forall pc rf st hp tgt src ma mb
    (STa: store pc rf st hp tgt src ma)
    (STb: store pc rf st hp tgt src mb)
    (VLDhp: pmvalid hp)
    (VLDst: pmvalid st),
  ma = mb.
Proof.
  i; inv STa; inv STb; try rename H1 into H0
  ; apply eq_sym, pmcell_uniq in H0; des; vauto.
Qed.





Theorem tgt_det: forall m evta evtb ma mb
    (STa: tgt_step evta m ma)
    (STb: tgt_step evtb m mb)
    (VLDhp: pmvalid (mhp m))
    (VLDst: pmvalid (mst m))
    (EVT: forall n, evta <> ein n),
  evta = evtb /\ ma = mb.
Proof.
  i; invc STa; invc STb
  ; [|by (exfalso; eapply ST0; eauto)|by (exfalso; eapply ST; eauto)].
  invc ST; invc ST0; hexploit (pmread_uniq INS INS0); intro ENC
  ; apply enc_inj in ENC; vauto.
  - hexploit (eval_uniq EVAL EVAL0); i; inst; subst; eauto.
  - rewrite ->EVALV in *; vauto.
  - rewrite ->EVALV in *; vauto.
  - rewrite ->EVALV in *; vauto.
    hexploit (eval_uniq EVALA EVALA0); i; inst; subst; eauto.
  - hexploit (eval_uniq EVAL EVAL0); i; inst; subst; eauto.
  - hexploit (store_uniq UPD UPD0); i; inst; subst; eauto.
  - hexploit (eval_uniq EVALa EVALa0); i; inst; subst; eauto.
    hexploit (eval_uniq EVALb EVALb0); i; inst; subst; eauto.
    hexploit (store_uniq UPD UPD0); i; inst; subst; eauto.
  - exfalso; eapply EVT; auto.
Qed.

Definition instr_wdet instr : Prop :=
  match instr with
    | i_inp _ | i_alc _ _ => False
    | _ => True
  end.

Lemma mkerr_error: forall m,
  error tgt_lsi (mkerr m).
Proof.
  i; destruct m; ii.
  rrdes FH; ss; rrdes HV; vauto.
Qed.

               
Lemma eval_frame: forall rf st stf hp hpf w w'
    (EVAL: eval rf st hp w w')
    (VLDs: pmvalid (st ## stf))
    (VLDh: pmvalid (hp ## hpf)),
  eval rf (st ## stf) (hp ## hpf) w w'.
Proof.
  i; inv EVAL; [econs 3|econs 4]; eapply pmread_plus; eauto.
Qed.

Definition mplus_sh m st hp :=
  {| mpc := mpc m; mrf := mrf m; mst := mst m ## st; mhp := mhp m ## hp |}.

               
Lemma store_frame: forall pc rf st stf hp hpf op w m
    (STO: store pc rf st hp op w m)
    (VLDs: pmvalid (st ## stf))
    (VLDh: pmvalid (hp ## hpf)),
  store pc rf (st ## stf) (hp ## hpf) op w (mplus_sh m stf hpf) /\
  pmvalid (mst m ## stf) /\ pmvalid (mhp m ## hpf) /\
  (pmfinite (hp ## hpf) -> pmfinite (mhp m ## hpf)).
Proof.
  i; inv STO; unfold mplus_sh; s; try rewrite <-!pmplus_assoc
  ; (splits; [by econs| | |]); eauto.
  - rewrite <-pmplus_assoc in VLDs; eapply pmvalid_plus_cell; eauto.
  - rewrite <-pmplus_assoc in VLDh; eapply pmvalid_plus_cell; eauto.
  - intro FIN; rewrite <-pmplus_assoc in VLDh; split_pmfinite FIN VLDh.
    repeat apply pmfinite_plus; eauto using pmfinite_cell.
Qed.

Lemma step_ok_frame: forall evt m inst m' cf mf
    (PC: pmread (mhp m) (mpc m) (enc inst))
    (DI: instr_wdet inst)
    (ST: step_ok evt m m')
    (PL: mcplus tgt_lsi m cf mf),
  step_ok evt mf (mplus_sh m' (cst cf) (chp cf)).
Proof.
  i; rrdes PL; destruct mf, cf, c'; inv ST
  ; unfold mplus_sh; rrdes RL RLm; ss; clarify
  ; try (by hexploit (pmread_uniq PC INS); intro E; apply enc_inj in E; clarify)
  ; eapply pmread_plus in INS; eauto
  ; destruct cpc0, crf0; ss; clarify
  ; eauto using Sjz, Slpc, Sout.
  - eapply eval_frame in EVAL; eauto using Sjmp.
  - eapply eval_frame in EVALA; eauto using Sjnz.
  - eapply eval_frame in EVAL; eauto using Sld.
  - eapply store_frame in UPD; des; eauto using Ssto.
  - eapply eval_frame in EVALa; eauto.
    eapply eval_frame in EVALb; eauto.
    eapply store_frame in UPD; des; eauto using Sopr.
Qed.

Lemma tgt_wdet: forall m inst
    (PC: pmread (mhp m) (mpc m) (enc inst))
    (DI: instr_wdet inst),
  @wdet tgt_lsi m.
Proof.
  ii; invc STa; [|by exfalso; eapply mkerr_error; eauto].
  invc STb; [|by exfalso; eapply ST0, step_ok_frame; eauto].
  rrdes PL; destruct inst; vauto; clear DI NZ
  ; (invc ST
    ; hexploit (pmread_uniq PC INS); intro E; apply enc_inj in E; clarify
    ; destruct cf; rrdes RLm; rrdes RL0; ss; clarify; clear INS
    ; rrdes RL; destruct ca', c'; ss; clarify
    ; destruct crf0, cpc0; vauto; ss; clarify
    ; inv ST0; ss; clarify; eapply pmread_plus in PC; eauto
    ; hexploit (pmread_uniq PC INS); intro E; apply enc_inj in E; clarify
    ; try (by omega)
    ; try (by split; auto; rr; splits; auto)).
  - eapply eval_frame in EVAL; eauto.
    hexploit (eval_uniq EVAL EVAL0); intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply eval_frame in EVALA; eauto.
    hexploit (eval_uniq EVALA EVALA0); intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply eval_frame in EVAL; eauto.
    hexploit (eval_uniq EVAL EVAL0); intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply store_frame in UPD; eauto; des.
    hexploit (store_uniq UPD UPD0); eauto; intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply eval_frame in EVALa; eauto.
    hexploit (eval_uniq EVALa EVALa0); intro E; instantiate; subst.
    eapply eval_frame in EVALb; eauto.
    hexploit (eval_uniq EVALb EVALb0); intro E; instantiate; subst.
    eapply store_frame in UPD; eauto; des.
    hexploit (store_uniq UPD UPD0); eauto; intro E; instantiate; subst.
    split; auto; rr; splits; auto.
Qed.

Lemma tgt_wfmod_nil: tgt_wfmod (fun _ _ => nil) nil.
Proof.
  rr; splits; vauto; i; apply nth_error_someD in INa; vauto.
Qed.
