(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import countable vgtac basis lang_common lang_spec lang_mid_renaming.
Require Import model_common model_untyped model_properties_untyped gw_untyped mm.
Require Import ssr util.
Require Import opt_transform lang_mid compiler_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


(** Merge consecutive let-bindings of the same expression. *)

(** * Definition. *)

Definition dedup: exptf :=
  fun e => match e with
    | IElet ea xa (IElet eb xb e') => 
      if eqdec ea eb then
        match xa, xb with
          | inl xa', inl xb' =>
            Some (IElet ea (inl xa') (rename ren_id[inr xa'/xb'] e'))
          | inr xa', inr xb' =>
            Some (IElet ea (inr xa') (rename ren_id[c xa'/xb'] e'))
          | _, _ =>
            None
        end
      else None
    | _ => None
  end.

Lemma dedup_bvars : forall (e e': exp)
    (E: dedup e = Some e'),
  sublist (bvars e') (bvars e).
Proof.
  i; destruct e; vauto; destruct e2, bx; vauto; destruct bx0; vauto;
  revert E; unfold dedup; destruct (eqdec _ _); try discriminate; 
  ii; clarify; s; rewrite ?app_assoc ?app_comm_cons; 
  eapply sublist_app; eauto; rewrite rename_bvars; eauto.
Qed.

(** * Syntactic correctness. *)

Lemma dedupP: forall {e e'}
    (E: dedup e = Some e'),
  exists a eb bx1 bx2 r,
      <<E : e = IElet a bx1 (IElet a bx2 eb) >>
   /\ <<E': e' = IElet a bx1 (rename r eb) >>
   /\ <<R : match bx1, bx2 with
            | inl bx1, inl bx2 => r = ren_id[inr bx1/bx2]
            | inr bx1, inr bx2 => r = ren_id[c bx1/bx2]
            | _, _ => False
            end>>.
Proof.
  i; dup E; destruct e; vauto; destruct e2, bx; vauto; destruct bx0; vauto;
  revert E0; unfold dedup; destruct (eqdec _ _); try discriminate; ii; clarify;
  esplits; try reflexivity.
Qed.

Lemma dedup_wfexp: forall e G e'
    (E: dedup e = Some e')
    (WF: wfexp G e)
    (UN: uniq (bvars e ++ purge_lbl G)),
  <<WF: wfexp G e'>> /\
  <<BV: subuniq (bvars e') (bvars e)>> /\
  <<GT: is_good_thunk e -> is_good_thunk e'>> /\
  <<GE: good_exp e -> good_exp e'>>.
Proof.
  i; hexploit dedup_bvars; eauto; []; i; destruct e; vauto.
  s in E; des_ifs; clarify.
  - s in WF; des; s; splits; auto using subuniq_sublist.
    + rewrite <-2(app_nil_l _); eapply wfexp_contract.
      s; eauto using rename_wfexp.
    + rr; vauto.
  - s in WF; des; s; splits; auto using subuniq_sublist.
    + rewrite <-2(app_nil_l _); eapply wfexp_contract.
      s; eauto using rename_wfexp.
    + rr; vauto.
Qed.


(** * Semantic correctness. *)

Lemma dedup_psim lw G ea eb: forall
    (E: dedup eb = Some ea)
    (WF: wfexp G eb)
    (UN: uniq (bvars eb ++ purge_lbl G))
    (SLW: stable lw),
  mm_psim lw G ren_id ea eb.
Proof.
  i; dup E; specialize (dedup_wfexp _ G E0 WF UN);
  specialize (dedupP E); ii; des; clarify.
  rr; splits; vauto; []; exists 0; i.
  eapply pesim_acquire; i.
  eapply pesim_real; i.
  
  rrdes HR. dup WHX. apply mm_confrel_shape in WHX. rrdes WHX.
  subst; clarify; dup RLa; dup RLb.
  rrdes RLa; rrdes RLb; inversion Em; inversion Em0.
  destruct hay, hby; destruct ee, ee0; inv Ec; inv Ec0; subst.
  clean; simpl cplus in *; noemp.
  destruct e, e0; s; clarify.
  rename e1 into enva, e3 into envb.
  
  assert (INxG: ~ In (tkvar2var bx1) G).
  { i; apply purge_lbl_tkvar2var in FH.
    eapply (uniq_appD _ _ UN); eauto.
    case bx1 as [x|kx]; s; auto.
  }
  assert (INyG: ~ In (tkvar2var bx2) G).
  { i; apply purge_lbl_tkvar2var in FH.
    eapply (uniq_appD _ _ UN); eauto.
    case bx2 as [y|ky], bx1 as [x|kx]; s; right; eapply in_or_app; right; s; auto.
  }

  assert (uniq ([tkvar2var bx1] ++ [tkvar2var bx2])). 
  { rewrite <-app_nil_r at 1; eapply (uniq_sublist UN), sublist_app
    ; auto using sublist_nil'.
    case bx1, bx2; s bvars; eapply sublist_cons_keep
    ; rewrite <-app_nil_l at 1; eapply sublist_app; auto using sublist_nil'.
  }

  assert (bx1 <> bx2) by (rr; i; subst; s in H; des; eauto).
  
  assert (In (tkvar2var bx1) (vars a) -> False).
  { s in WF; s in UN; destruct bx1, bx2; des; i; eapply INxG; eapply wfexp_vars; eauto; 
    i; eapply UN; eauto using in_or_app. 
  }

  destruct (@mm_psim_rename lw eb0 (tkvar2var bx2 :: tkvar2var bx1 :: G) r); des; eauto.
  { destruct bx1, bx2; s in WF; des; eauto. }
  { intros ? ?; destruct bx1, bx2; subst; vauto.
    - apply in_consD in IN; des; subst.
      + s; tmext_simpl; i; exfalso.
        s in UN; des; eapply UN; rewrite cons_app; eauto 20 using in_or_app.
      + rewrite ren_ext_neq ?ren_var_id; vauto.
    - apply in_consD in IN; des; subst.
      + s; tmext_simpl; i; exfalso.
        s in UN; des; eapply UN; rewrite cons_app; eauto 20 using in_or_app.
      + rewrite ren_c_ext_neq ?ren_var_id; vauto.
  }

  destruct bx1 as [bx1|bx1], bx2 as [bx2|bx2]; s in R; try by exfalso; auto. 
  - (*| two regular lets *)

    destruct (LEM (exists va, eval envb a = Some va)) as [Hva|Hva]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst. esplits.
      econs 1; inv CNF; des; subst; rrdes MB; inv Ec; invc Ec1; first econs 1.
      - eapply IRerr; i; first by auto. inv FH. apply Hva. by eauto using EVAL.
      - eauto.
    }

    destruct Hva as [va Eva].
    
    destruct (LEM (exists va2, eval ((inl (inr bx1), va)::envb) a = Some va2)) as [Hva2|Hva2]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst. econs.
      - invc CNF. des; subst; rrdes MB; inv Ec; inv Ec1; econs 3; inst_pairs.
        + econs 1. econs. econs. by eauto.
        + econs 1. eapply IRerr; i; first by auto.
          inv FH. apply Hva2. by eauto using EVAL.
      - eauto. 
    }
    destruct Hva2 as [va2 Eva2].

    assert (Hvb': exists va', <<EVa' : eval enva a = Some va' >>
                           /\ <<VCL2 : vclos (wlift lw) s (g s) va' va2 >>).
    { eapply (mm_asim_refl _ G); last exact Eva2; eauto.
      - s in WF; des; eauto.
      - i; rewrite list_assoc_cons_neq; eauto.
        ii; clarify.
    }
    assert (Hvb'': exists va', <<EVa' : eval enva a = Some va' >>
                           /\ <<VCL' : vclos (wlift lw) s (g s) va' va >>).
    { s in WF; des; eauto; eapply (mm_asim_refl _ G); last exact Eva; eauto. }
    des; clarify.
    
    eapply pesim_dtsteps.
    + eassumption.
    + econs 2; first eapply mid_wdet.
      * inst_pairs; econs; econs; by eauto.
      * inst_pairs. econs 1; econs; by eauto.
      * econs 1.
    + i; econs. esplits; [| |exact PV| exact FIN]; inst_pairs; last reflexivity; eauto.
    + eauto.
    + i; eexists; rrdes RLb. 
      destruct XTR as [Xh]; subst. inv Ec1; clean.
      esplits; first 1 last. 
      * left; econs 2; econs 1; inst_pairs; econs 1; eapply IRlet; eauto.
      * rr; esplits; inst_pairs; s; noemp; eauto.
      * rewrite mid_csplit mid_csplit.
        eapply pesim_release; eauto; try by (rr; esplits; eauto).
        eapply H4; eauto.
        {
          - ii; s in IN; clarify; des; eauto.
            + injection IN; ii; clarify. s rename_itvar; rewrite tmext_eq.
              rewrite !list_assoc_cons_eq. eauto 7.
            + injection IN; ii; clarify. s rename_itvar. rewrite tmext_neq; ii; clarify. eauto.
              rewrite list_assoc_cons_eq. rewrite list_assoc_cons_neq; ii; clarify. eauto.
              rewrite list_assoc_cons_eq. eauto 7.
            + destruct x; ii; clarify; rewrite !list_assoc_cons_neq; ii; clarify; eauto.
              * rewrite (_ : (rename_itvar _ _ = rename_itvar ren_id (inl l))); eauto 7.
              * rewrite (_ : (rename_itvar _ _ = rename_itvar ren_id (inr t))); eauto 7.
                s. rewrite tmext_neq; ii; clarify; eauto.
              * rewrite tmext_neq in H3; ii; clarify; eauto.
        }
        { s; ii; des; clarify; eauto. by eapply RK. }
    
  - (*| two continuation lets *)

    destruct (LEM (exists ax e', a = IEcont ax e')) as [Hva|Hva]; last first.
    { i. pfold. econs 1. i. destruct XTRa as [Xha], XTRb as [Xhb]; subst. econs.
      - econs 1. inv CNF; des; subst; rrdes MB; inv Ec1.
        eapply IRerr; i; first by auto. inv FH. apply Hva. by eauto. 
      - eauto.
    }
    rrdes Hva; clarify.
    
    eapply pesim_dtsteps.
    + eassumption.
    + econs 2; first by eapply mid_wdet.
      * inst_pairs. econs. by econs.
      * econs 1.
    + i; econs. esplits; [| |exact PV| exact FIN]; inst_pairs; last reflexivity; eauto.
    + eauto.
    + i; eexists; rrdes RLb. 
      destruct XTR as [Xh]; subst. inv Ec1; clean.
      esplits; first 1 last. 
      * left; econs 2; econs 1; inst_pairs; econs 1; econs; eauto.
      * rr; esplits; inst_pairs; s; noemp; eauto.
      * rewrite mid_csplit mid_csplit.
        eapply pesim_release; eauto; try by (rr; esplits; eauto).
        eapply H4; eauto.
        { s; ii; des; clarify; eauto. }
        {
          - ii; s in IN; clarify; des; eauto.
            + injection IN; ii; clarify. s rename_contvar; rewrite tmext_eq.
              rewrite !list_assoc_cons_eq. esplits; eauto.
              assert (HK : mm_ksim lw G ren_id (IEcont ax e') (IEcont ax e')).
              { eapply mm_ksim_refl; auto.
                s in WF; des; eauto.
              }
              edestruct (HK); des; [by eauto|by eauto| | |by eauto|by eauto]; eauto.
              i; rewrite list_assoc_cons_neq; eauto; ii; clarify.
            + injection IN; ii. rewrite <- H3; clear x IN H3.
              s rename_contvar. rewrite tmext_neq; last ii; clarify; eauto.
              rewrite list_assoc_cons_eq.
              rewrite list_assoc_cons_neq; last ii; clarify; eauto.
              rewrite list_assoc_cons_eq. esplits; eauto.
              assert (HK : mm_ksim lw G ren_id (IEcont ax e') (IEcont ax e')).
              { eapply mm_ksim_refl; auto.
                s in WF; des; eauto.
              }
              edestruct (HK); des; [by eauto|by eauto| | |by eauto|by eauto]; eauto.
            + rewrite !list_assoc_cons_neq; ii; clarify; eauto.
              rewrite (_ : rename_contvar _ _ = rename_contvar ren_id x); eauto.
              * s; rewrite tmext_neq; ii; clarify; eauto.
              * rewrite tmext_neq in H3; ii; clarify; eauto.
        }
Qed.


Definition dedup_ann: annT.
  refine {| _annT := unit
          ; perform u := dedup
         |}.
  - ii; eauto using dedup_wfexp.
  - ii; econs; eapply dedup_psim; eauto.
Defined.
  
Definition analyze: exp -> aexp dedup_ann :=
  annotate_with tt.

Definition dedup_mod: modtf :=
  fun _ => transform_mod analyze.

Theorem dedup_wfmod: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  mid_wfmod FI (dedup_mod FI m) FO /\
  uniqvar (dedup_mod FI m).
Proof.
  i; eapply transform_wfmod; eauto using annotate_with_erase.
Qed.

Theorem dedup_msim: forall FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m),
  rtc (msim mm_gw FI FO) (dedup_mod FI m) m.
Proof.
  i; eapply transform_msim; eauto using annotate_with_erase.
Qed.  

Theorem dedup_correct:
  <<WF: modtf_wf dedup_mod>> /\
  <<SIM: modtf_sim dedup_mod>>.
Proof.
  split; ii; eauto using dedup_wfmod, dedup_msim.
Qed.
